#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os
import os.path
import shlex
import subprocess
import sys

import pytest

import intake
from intake.source import discovery


@pytest.fixture
def extra_pythonpath():
    basedir = os.path.dirname(__file__)
    extra_path = os.path.join(basedir, 'plugin_searchpath')

    # Put extra directory on the python path
    sys.path.append(extra_path)

    yield extra_path

    # Return python path back to normal
    sys.path.remove(extra_path)


def test_package_scan(extra_pythonpath, tmp_config_path):
    "This tests a non-public function."
    # Default path (sys.path)
    results = discovery._package_scan()
    assert 'foo' in results

    # Explicit path
    results = discovery._package_scan(path=[extra_pythonpath])
    assert 'foo' in results


def test_discover_cli(extra_pythonpath, tmp_config_path):
    env = os.environ.copy()
    env["INTAKE_CONF_FILE"] = tmp_config_path
    env['PYTHONPATH'] = extra_pythonpath

    # directory is not automatically scanned any more
    subprocess.call(shlex.split(
        "intake drivers enable foo intake_foo.FooPlugin"
    ), stderr=subprocess.STDOUT, stdout=subprocess.PIPE, env=env)

    out = subprocess.check_output(shlex.split(
        "intake drivers list"
    ), stderr=subprocess.STDOUT, env=env)

    assert b'foo' in out
    assert out.index(b'Disabled') > out.index(b'foo')

    subprocess.check_output(shlex.split(
        "intake drivers disable foo"
    ), stderr=subprocess.STDOUT, env=env)

    out = subprocess.check_output(shlex.split(
        "intake drivers list"
    ), stderr=subprocess.STDOUT, env=env)
    assert b'foo' in out
    assert out.index(b'Disabled') < out.index(b'foo')


def test_discover(extra_pythonpath, tmp_config_path):
    drivers = intake.source.discovery.DriverSouces(do_scan=True)
    with pytest.warns(PendingDeprecationWarning):
        assert "foo" in drivers.scanned

    registry = intake.source.DriverRegistry(drivers)
    # Check that package scan (name-based) discovery worked.
    assert 'foo' in registry
    registry['foo']()
    # Check that entrypoints-based discovery worked.
    assert 'some_test_driver' in registry
    registry['some_test_driver']()

    # Now again, turning off the package scan.

    drivers = intake.source.discovery.DriverSouces()
    registry = intake.source.DriverRegistry(drivers)

    # Check that package scan (name-based) discovery did *not* happen.
    assert 'foo' not in registry
    # Check that entrypoints-based discovery worked.
    assert 'some_test_driver' in registry
    registry['some_test_driver']()


def test_enable_and_disable(extra_pythonpath, tmp_config_path):
    # Disable and then enable a package scan result.

    try:
        drivers = intake.source.discovery.DriverSouces(do_scan=True)
        drivers.disable('foo')
        registry = intake.source.DriverRegistry(drivers)
        with pytest.warns(PendingDeprecationWarning):
            assert 'foo' in discovery.drivers.scanned
        assert 'foo' not in registry

        drivers.enable('foo', 'intake_foo.FooPlugin')
        assert 'foo' in registry
    finally:
        drivers.enable('foo', 'intake_foo.FooPlugin')

    # Disable and then enable an entrypoint result.

    try:
        drivers.disable('some_test_driver')
        assert 'some_test_driver' not in registry

        drivers.enable(
            'some_test_driver',
            'driver_with_entrypoints.SomeTestDriver')
        assert 'some_test_driver' in registry
    finally:
        drivers.enable(
            'some_test_driver',
            'driver_with_entrypoints.SomeTestDriver')


def test_discover_collision(extra_pythonpath, tmp_config_path):
    with pytest.warns(UserWarning):
        discovery._package_scan(plugin_prefix='collision_')



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os
import os.path
import shutil
import tempfile
import time

import pytest

from intake import open_catalog

from intake.auth.secret import SecretClientAuth
from intake.auth.base import AuthenticationFailure

TMP_DIR = tempfile.mkdtemp()
CONF_DIR = os.path.join(TMP_DIR, 'conf')
os.mkdir(CONF_DIR)

TEST_CATALOG_PATH = [TMP_DIR]
YAML_FILENAME = 'intake_test_catalog.yml'


# Create server configuration using shared-secret Auth
TEST_SERVER_CONF = os.path.join(CONF_DIR, 'config.yaml')
conf = '''
auth:
  cls: intake.auth.secret.SecretAuth
  kwargs:
    secret: test_secret
'''
with open(TEST_SERVER_CONF, 'w') as f:
    f.write(conf)


@pytest.fixture
def intake_server_with_auth(intake_server):
    fullname = os.path.join(TMP_DIR, YAML_FILENAME)

    try:
        os.makedirs(os.path.join(TMP_DIR, 'data'))
    except:
        pass
    with open(fullname, 'w') as f:
        f.write('''
sources:
  example:
    description: example1 source plugin
    driver: csv
    args:
      urlpath: "{{ CATALOG_DIR }}/data/example.csv"
        ''')

    csv_name = os.path.join(TMP_DIR, 'data', 'example.csv')
    with open(csv_name, 'w') as f:
        f.write('a,b,c\n1,2,3\n4,5,6')
    time.sleep(2)

    yield intake_server

    try:
        shutil.rmtree(TMP_DIR)
    except:
        pass


def test_secret_auth(intake_server_with_auth):
    auth = SecretClientAuth(secret='test_secret')
    catalog = open_catalog(intake_server_with_auth, auth=auth)

    entries = list(catalog)
    assert entries == ['example']

    catalog.example.read()


def test_secret_auth_fail(intake_server_with_auth):
    auth = SecretClientAuth(secret='test_wrong_secret')
    with pytest.raises(AuthenticationFailure):
        list(open_catalog(intake_server_with_auth, auth=auth))



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import intake
import os

here = os.path.abspath(os.path.dirname(__file__))
fn = os.path.join(here, 'catalog_alias.yml')


def test_simple():
    cat = intake.open_catalog(fn)
    s = cat.alias0()
    assert s.container == 'other'
    out = str(s.discover())
    assert s.container == 'dataframe'
    assert "state" in out


def test_mapping():
    cat = intake.open_catalog(fn)
    s = cat.alias1()
    assert s.container == 'other'
    out = str(s.discover())
    assert s.container == 'dataframe'
    assert "state" in out

    s = cat.alias1(choice='second')
    assert s.container == 'other'
    out = str(s.discover())
    assert s.container == 'ndarray'
    assert "int64" in out



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
import pytest
pytest.importorskip('panel')


@pytest.fixture
def search_inputs(cat1, cat2):
    from ..search import SearchInputs
    return SearchInputs()


def test_search_inputs(search_inputs):
    assert search_inputs.visible
    assert len(search_inputs.children) == 4
    assert len(search_inputs.panel.objects) == 4


def test_search_inputs_text_prop_equal_to_widget_value(search_inputs):
    search_inputs.text = 'some text'
    assert search_inputs.text_widget.value == 'some text'


def test_search_inputs_depth_prop_parses_to_int(search_inputs):
    search_inputs.depth = '2'
    assert search_inputs.depth == 2

    search_inputs.depth = 'All'
    assert search_inputs.depth == 99


@pytest.fixture
def search(cat1, cat2):
    from ..search import Search
    return Search(cats=[cat1, cat2])


def test_search(search):
    assert search.visible
    assert len(search.children) == 2
    assert len(search.panel.objects) == 2


def test_search_watchers_gets_populated(search):
    assert len(search.watchers) == 1


def test_search_widget_click_tries_to_run_callback(search):
    search.inputs.text = 'flight'
    with pytest.raises(TypeError, match="'NoneType' object is not callable"):
        search.widget.clicks = 1


def test_search_unwatch_watchers_get_cleaned_up(search):
    search.unwatch()
    assert len(search.watchers) == 0
    search.inputs.text = 'flight'

    # does not try to run callback
    search.widget.clicks = 2


def test_callback_gets_right_input(search):
    def callback(new_cats):
        """Raises an error if called"""
        raise ValueError('New catalogs', new_cats)

    search.inputs.text = 'flight'
    search.done_callback = callback
    with pytest.raises(ValueError, match='<Intake catalog: catalog1_search>'):
        search.widget.clicks = 3



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
import pytest
import intake.catalog.utils as utils
import pandas as pd


def test_expand_templates():
    pars = {'a': "{{par}} hi"}
    context = {'b': 1, 'par': 'ho'}
    assert utils.expand_templates(pars, context)['a'] == 'ho hi'
    assert utils.expand_templates(pars, context, True)[1] == {'b'}


def test_expand_nested_template():
    pars = {'a': ["{{par}} hi"]}
    context = {'b': 1, 'par': 'ho'}
    assert utils.expand_templates(pars, context)['a'] == ['ho hi']
    assert utils.expand_templates(pars, context, True)[1] == {'b'}

    pars = {'a': {'k': {("{{par}} hi", )}}}
    context = {'b': 1, 'par': 'ho'}
    assert utils.expand_templates(pars, context)['a'] == {'k': {("ho hi", )}}
    assert utils.expand_templates(pars, context, True)[1] == {'b'}


@pytest.mark.parametrize("test_input,expected", [
   (None, pd.Timestamp('1970-01-01 00:00:00')),
   (1, pd.Timestamp('1970-01-01 00:00:00.000000001')),
   ("1988-02-24T13:37+0100", pd.Timestamp("1988-02-24 13:37+0100")),
   ({"__datetime__": True, "as_str": "1988-02-24T13:37+0100"}, pd.Timestamp("1988-02-24T13:37+0100")),
])
def test_coerce_datetime(test_input, expected):
    assert utils.coerce_datetime(test_input) == expected


def test_flatten():
    assert list(utils.flatten([["hi"], ["oi"]])) == ['hi', 'oi']

@pytest.mark.parametrize(
    "value,dtype,expected", [
        (1, "int", 1),
        ("1", "int", 1),
        (1, "str", "1"),
        ((), "list", []),
        ((1, ), "list", [1]),
        ((1, ), "list[str]", ["1"])
    ],
)
def test_coerce(value, dtype, expected):
    out = utils.coerce(dtype, value)
    assert out == expected
    assert type(out) == type(expected)



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
import os
import pytest
pn = pytest.importorskip('panel')

def callback(args):
    """Raises an error if called"""
    raise ValueError('Callback provided:', args)


@pytest.fixture
def file_selector():
    from ..add import FileSelector
    return FileSelector()


@pytest.fixture
def url_selector():
    from ..add import URLSelector
    return URLSelector()


@pytest.fixture
def cat_adder():
    from ..add import CatAdder
    return CatAdder()


def test_file_selector(file_selector):
    assert file_selector.path == os.getcwd() + '/'


def test_file_selector_raises_error_if_no_file_selected(file_selector):
    with pytest.raises(IndexError, match='list index out of range'):
        file_selector.url


def test_file_selector_edit_path(file_selector):
    expected = os.getcwd()
    file_selector.move_up()
    file_selector.path_text.value = os.getcwd()
    assert file_selector.validator.object is None
    assert file_selector.path == expected


def test_file_selector_go_home(file_selector):
    expected = os.getcwd() + '/'
    assert file_selector.path == expected
    file_selector.move_up()
    file_selector.go_home()
    assert file_selector.path == expected


def test_file_selector_move_up(file_selector):
    assert file_selector.path == os.getcwd() + '/'
    file_selector.move_up()
    expected = os.path.abspath('..')
    assert file_selector.path == expected


def test_file_selector_move_down(file_selector):
    expected = os.getcwd() + '/'
    dirname = expected.split('/')[-2] + '/'

    # move up so that we know we will be able to move down into
    # intial dir
    file_selector.move_up()

    # setting the value on main widget will trigger move down
    file_selector.main.value = [dirname]
    assert file_selector.path == expected

    # should empty the selection on main
    assert file_selector.main.value == []

def test_url_selector(url_selector):
    assert url_selector.url == ''
    assert url_selector.visible
    assert len(url_selector.panel.objects) == 2

def test_url_selector_set_visible_to_false(url_selector):
    url_selector.visible = False
    assert url_selector.visible is False
    assert len(url_selector.panel.objects) == 0


def test_cat_adder(cat_adder):
    assert cat_adder.visible is True
    assert cat_adder.tabs.active == 0
    assert cat_adder.widget.disabled is True
    assert len(cat_adder.panel.objects) == 2

    cat_adder.tabs.active = 1
    assert cat_adder.cat_url[0] == ''
    assert cat_adder.cat.name is None
    assert cat_adder.widget.disabled is False

    cat_adder.done_callback = callback
    with pytest.raises(ValueError, match='None'):
        cat_adder.add_cat()

def test_cat_adder_add_real_cat(cat_adder, cat1_url, cat1):
    cat_adder.tabs.active = 1
    cat_adder.url.main.value = cat1_url

    assert cat_adder.cat_url[0] == cat1_url
    assert cat_adder.cat == cat1

    cat_adder.done_callback = callback
    with pytest.raises(ValueError, match=str(cat1)):
        cat_adder.add_cat()



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
import pytest
pn = pytest.importorskip('panel')


@pytest.fixture
def gui(cat1, cat2):
    from ..gui import CatGUI
    return CatGUI(cats=[cat1, cat2])


def test_gui(gui, cat1, cat2):
    assert gui.select.items == [cat1, cat2]
    assert gui.cats == [cat1]

    assert not gui.add.watchers
    assert gui.add.visible is False
    assert gui.add_widget.disabled is False

    assert not gui.search.watchers
    assert gui.search.visible is False
    assert gui.search_widget.disabled is False


def test_gui_close_and_open_select(gui, cat2, sources2):
    gui.select.selected = [cat2]
    gui.select.visible = False
    assert not gui.select.watchers
    assert gui.search_widget.disabled is False

    gui.select.visible = True
    assert len(gui.select.watchers) == 1
    assert gui.select.selected == [cat2]
    assert gui.search_widget.disabled is False


def test_gui_getstate(gui, cat1, sources1):
    state = gui.__getstate__()

    assert state['visible']is True
    assert state['add']['visible'] is False
    assert state['search']['visible'] is False
    assert state['select']['selected'] == [cat1.name]


def test_gui_state_roundtrip(gui, cat1, cat2, sources1):
    from ..gui import CatGUI
    other = CatGUI.from_state(gui.__getstate__())

    assert other.select.items == [cat1, cat2]
    assert other.cats == [cat1]
    assert other.search.visible is False
    assert other.add.visible is False


def test_gui_state_roundtrip_with_subpanels(gui, cat1, cat2, sources1):
    from ..gui import CatGUI
    gui.search.visible = True
    gui.search.inputs.text = 'foo'
    gui.search.inputs.depth = 3
    gui.add.visible = True
    gui.add.tabs.active = 1

    other = CatGUI.from_state(gui.__getstate__())

    assert other.select.items == [cat1, cat2]
    assert other.cats == [cat1]
    assert other.search.visible is True
    assert other.search_widget.value is True
    assert other.search.inputs.text == 'foo'
    assert other.search.inputs.depth == 3

    assert other.add.visible is True
    assert other.add_widget.value is True
    assert other.add.widget.disabled is False



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from pathlib import Path
import sys
from intake.catalog import default
from intake.catalog.base import Catalog


def test_which():
    p = default.which('python')
    assert Path(p).resolve() == Path(sys.executable).resolve()


def test_load():
    cat = default.load_user_catalog()
    assert isinstance(cat, Catalog)
    cat = default.load_global_catalog()
    assert isinstance(cat, Catalog)



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import intake
import pytest

pytest.importorskip('panel')


def assert_widget_matches(browser):
    assert browser.options == browser.widget.options
    assert browser.selected == browser.widget.value


def test_catalog_browser_init_emtpy():
    from ..select import CatSelector
    cat_browser = CatSelector()
    assert cat_browser.selected == [intake.cat]
    assert_widget_matches(cat_browser)


def test_catalog_browser(cat_browser, cat1):
    assert cat_browser.items == [cat1]
    assert cat1.name in cat_browser.options
    assert cat_browser.selected == [cat1]
    assert_widget_matches(cat_browser)


def test_catalog_browser_set_to_visible_and_back(cat_browser, cat1):
    cat_browser.visible = False
    assert len(cat_browser.watchers) == 0

    cat_browser.visible = True
    assert len(cat_browser.watchers) == 1
    assert cat_browser.items == [cat1]
    assert cat1.name in cat_browser.options
    assert cat_browser.selected == [cat1]
    assert_widget_matches(cat_browser)


def test_catalog_browser_add(cat_browser, cat2):
    cat_browser.add(cat2)
    assert cat2.name in cat_browser.options
    assert cat_browser.selected == [cat2]
    assert_widget_matches(cat_browser)


def test_catalog_browser_add_list(cat_browser, cat2):
    cat_browser.add([cat2])
    assert cat2.name in cat_browser.options
    assert cat_browser.selected == [cat2]
    assert_widget_matches(cat_browser)


def test_catalog_browser_add_cat_as_str(cat_browser, cat2, cat2_url):
    cat_browser.add(cat2_url)
    assert cat2.name in cat_browser.options
    assert cat_browser.selected == [cat2]
    assert_widget_matches(cat_browser)


def test_catalog_browser_add_nested_catalog(cat_browser, parent_cat):
    cat_browser.add(parent_cat)
    assert parent_cat.name in cat_browser.options
    assert cat_browser.selected == [parent_cat]
    assert list(cat_browser.options.keys()) == ['catalog1', 'parent', '└── child1', '└── child2']
    assert_widget_matches(cat_browser)


def test_catalog_browser_select_cat_by_widget(cat_browser, cat1):
    cat_browser.selected = []
    assert cat_browser.selected == []
    assert_widget_matches(cat_browser)

    cat_browser.widget.value = [cat1]
    assert cat_browser.selected == [cat1]
    assert_widget_matches(cat_browser)


def test_catalog_browser_remove_selected_cat(cat_browser, cat1):
    cat_browser.remove_selected()
    assert cat1 not in cat_browser.options
    assert cat_browser.selected == []
    assert_widget_matches(cat_browser)


def test_catalog_browser_remove_cat_that_is_not_in_options_passes(cat_browser, cat2):
    assert cat2.name not in cat_browser.options
    cat_browser.remove(cat2)


def test_catalog_browser_remove_nested_catalog(cat_browser, parent_cat):
    cat_browser.add(parent_cat)
    assert parent_cat.name in cat_browser.options
    assert cat_browser.selected == [parent_cat]
    assert list(cat_browser.options.keys()) == ['catalog1', 'parent', '└── child1', '└── child2']
    cat_browser.remove_selected()
    assert list(cat_browser.options.keys()) == ['catalog1']
    assert_widget_matches(cat_browser)



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------


def verify_plugin_interface(plugin):
    assert isinstance(plugin.version, str)
    assert isinstance(plugin.container, str)
    assert isinstance(plugin.partition_access, bool)


def verify_datasource_interface(source):
    for attr in ['container', 'description', 'dtype', 'shape',
                 'npartitions', 'metadata']:
        assert hasattr(source, attr)

    for method in ['discover', 'read', 'read_chunked', 'read_partition',
                   'to_dask', 'close']:
        assert hasattr(source, method)



import os
import pytest
import intake
from intake.catalog.local import LocalCatalogEntry, UserParameter
from intake.source.base import DataSource


class NoSource(DataSource):

    def __init__(self, **kwargs):
        self.metadata = kwargs.pop('metadata', {})
        self.kwargs = kwargs


driver = 'intake.catalog.tests.test_parameters.NoSource'


def test_simplest():
    e = LocalCatalogEntry('', '', driver, args={'arg1': 1})
    s = e()
    assert s.kwargs['arg1'] == 1


def test_cache_default_source():
    # If the user provides parameters, don't allow default caching
    up = UserParameter('name', default='oi')
    e = LocalCatalogEntry('', '', driver, getenv=False, parameters=[up])
    s1 = e(name="oioi")
    s2 = e()
    assert s1 is not s2
    s1 = e()
    s2 = e(name="oioi")
    assert s1 is not s2
    # Otherwise, we can cache the default source
    e = LocalCatalogEntry('', '', driver, getenv=False)
    s1 = e()
    s2 = e()
    assert s1 is s2


def test_parameter_default():
    up = UserParameter('name', default='oi')
    e = LocalCatalogEntry('', '', driver, args={'arg1': "{{name}}"},
                          parameters=[up])
    s = e()
    assert s.kwargs['arg1'] == 'oi'


def test_maybe_default_from_env():
    # maybe fill in parameter default from the env, depending on getenv
    up = UserParameter('name', default='env(INTAKE_TEST_VAR)')
    e = LocalCatalogEntry('', '', driver, args={'arg1': "{{name}}"},
                          parameters=[up], getenv=False)
    s = e()
    assert s.kwargs['arg1'] == 'env(INTAKE_TEST_VAR)'

    os.environ['INTAKE_TEST_VAR'] = 'oi'
    # Clear the cached source so we can (not) pick up the changed environment variable.
    e.clear_cached_default_source()

    s = e()
    assert s.kwargs['arg1'] == 'env(INTAKE_TEST_VAR)'

    up = UserParameter('name', default='env(INTAKE_TEST_VAR)')
    e = LocalCatalogEntry('', '', driver, args={'arg1': "{{name}}"},
                          parameters=[up], getenv=True)
    s = e()
    assert s.kwargs['arg1'] == 'oi'

    del os.environ['INTAKE_TEST_VAR']
    # Clear the cached source so we can pick up the changed environment variable.
    e.clear_cached_default_source()

    s = e()
    assert s.kwargs['arg1'] == ''


def test_up_override_and_render():
    up = UserParameter('name', default='env(INTAKE_TEST_VAR)')
    e = LocalCatalogEntry('', '', driver, args={'arg1': "{{name}}"},
                          parameters=[up], getenv=False)
    s = e(name='other')
    assert s.kwargs['arg1'] == 'other'


def test_user_explicit_override():
    up = UserParameter('name', default='env(INTAKE_TEST_VAR)')
    e = LocalCatalogEntry('', '', driver, args={'arg1': "{{name}}"},
                          parameters=[up], getenv=False)
    # override wins over up
    s = e(arg1='other')
    assert s.kwargs['arg1'] == 'other'


def test_auto_env_expansion():
    os.environ['INTAKE_TEST_VAR'] = 'oi'
    e = LocalCatalogEntry('', '', driver,
                          args={'arg1': "{{env(INTAKE_TEST_VAR)}}"},
                          parameters=[], getenv=False)
    s = e()

    # when getenv is False, you pass through the text
    assert s.kwargs['arg1'] == '{{env(INTAKE_TEST_VAR)}}'

    e = LocalCatalogEntry('', '', driver,
                          args={'arg1': "{{env(INTAKE_TEST_VAR)}}"},
                          parameters=[], getenv=True)
    s = e()
    assert s.kwargs['arg1'] == 'oi'

    # same, but with quoted environment name
    e = LocalCatalogEntry('', '', driver,
                          args={'arg1': '{{env("INTAKE_TEST_VAR")}}'},
                          parameters=[], getenv=True)
    s = e()
    assert s.kwargs['arg1'] == 'oi'

    del os.environ['INTAKE_TEST_VAR']
    # Clear the cached source so we can pick up the changed environment variable.
    e.clear_cached_default_source()

    s = e()
    assert s.kwargs['arg1'] == ''

    e = LocalCatalogEntry('', '', driver,
                          args={'arg1': "{{env(INTAKE_TEST_VAR)}}"},
                          parameters=[], getenv=False)
    s = e()
    assert s.kwargs['arg1'] == '{{env(INTAKE_TEST_VAR)}}'


def test_validate_up():
    up = UserParameter('name', default=1, type='int')
    e = LocalCatalogEntry('', '', driver, args={'arg1': "{{name}}"},
                          parameters=[up], getenv=False)
    s = e()  # OK
    assert s.kwargs['arg1'] == '1'
    with pytest.raises(ValueError):
        e(name='oi')

    up = UserParameter('name', type='int')
    e = LocalCatalogEntry('', '', driver, args={'arg1': "{{name}}"},
                          parameters=[up], getenv=False)
    s = e()  # OK
    # arg1 is a string: real int gets rendered by jinja
    assert s.kwargs['arg1'] == '0'  # default default for int
    s = e(arg1='something')
    assert s.kwargs['arg1'] == 'something'


def test_validate_par():
    up = UserParameter('arg1', type='int')
    e = LocalCatalogEntry('', '', driver, args={'arg1': "oi"},
                          parameters=[up], getenv=False)
    with pytest.raises(ValueError):
        e()
    e = LocalCatalogEntry('', '', driver, args={'arg1': 1},
                          parameters=[up], getenv=False)
    e()  # OK

    e = LocalCatalogEntry('', '', driver, args={'arg1': "1"},
                          parameters=[up], getenv=False)
    s = e()  # OK
    assert s.kwargs['arg1'] == 1  # a number, not str


def test_mlist_parameter():
    up = UserParameter("", "", "mlist", allowed=["a", "b"])
    up.validate([])
    up.validate(['b'])
    up.validate(['b', 'a'])
    with pytest.raises(ValueError):
        up.validate(["c"])
    with pytest.raises(ValueError):
        up.validate(["a", "c"])
    with pytest.raises(ValueError):
        up.validate("hello")


def test_explicit_overrides():
    e = LocalCatalogEntry('', '', driver, args={'arg1': "oi"})
    s = e(arg1='hi')
    assert s.kwargs['arg1'] == 'hi'

    e = LocalCatalogEntry('', '', driver, args={'arg1': "{{name}}"})
    s = e(name='hi')
    assert s.kwargs['arg1'] == 'hi'

    os.environ['INTAKE_TEST_VAR'] = 'another'
    e = LocalCatalogEntry('', '', driver, args={'arg1': "oi"}, getenv=True)
    s = e(arg1='{{env(INTAKE_TEST_VAR)}}')
    assert s.kwargs['arg1'] == 'another'

    up = UserParameter('arg1', type='int')
    e = LocalCatalogEntry('', '', driver, args={'arg1': 1},
                          parameters=[up])
    with pytest.raises(ValueError):
        e(arg1='oi')

    s = e(arg1='1')
    assert s.kwargs['arg1'] == 1


def test_extra_arg():
    e = LocalCatalogEntry('', '', driver, args={'arg1': "oi"})
    s = e(arg2='extra')
    assert s.kwargs['arg2'] == 'extra'


def test_unknown():
    e = LocalCatalogEntry('', '', driver, args={'arg1': "{{name}}"})
    s = e()
    assert s.kwargs['arg1'] == ""

    # parameter has no default
    up = UserParameter('name')
    e = LocalCatalogEntry('', '', driver, args={'arg1': "{{name}}"},
                          parameters=[up])
    s = e()
    assert s.kwargs['arg1'] == ""


def test_catalog_passthrough():
    root = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
    cat = intake.open_catalog(os.path.join(root, "tests/catalog_inherit_params.yml"))
    assert set(cat.subcat.user_parameters) == {"bucket", "inner"}
    url = cat.subcat.ex2.urlpath
    assert url == "test_bucket/test_name"
    url = cat.subcat.ex2(bucket="hi", inner="ho").urlpath
    assert url == "hi/ho"

    # test clone
    cat2 = cat(bucket="yet")
    url = cat2.subcat(inner="another").ex2.urlpath
    assert url == "yet/another"
    url = cat2.subcat(inner="another").ex2(bucket="hi").urlpath
    assert url == "hi/another"



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os
import pytest
import intake
from intake.source.textfiles import TextFilesSource
from intake.source import import_name
from fsspec import open_files
here = os.path.abspath(os.path.dirname(__file__))


def test_textfiles(tempdir):
    open(os.path.join(tempdir, '1.txt'), 'wt').write('hello\nworld')
    open(os.path.join(tempdir, '2.txt'), 'wt').write('hello\nworld')
    path = os.path.join(tempdir, '*.txt')
    t = TextFilesSource(path)
    t.discover()
    assert t.npartitions == 2
    assert t._get_partition(0) == t.to_dask().to_delayed()[0].compute()
    out = t.read()
    assert isinstance(out, list)
    assert out[0] == 'hello\n'


@pytest.mark.parametrize('comp', [None, 'gzip', 'bz2'])
def test_complex_text(tempdir, comp):
    dump, load, read = 'json.dumps', 'json.loads', True
    dump = import_name(dump)
    data = [{'something': 'simple', 'and': 0}] * 2
    for f in ['1.out', '2.out']:
        fn = os.path.join(tempdir, f)
        with open_files([fn], mode='wt', compression=comp)[0] as fo:
            if read:
                fo.write(dump(data))
            else:
                dump(data, fo)
    # that was all setup

    path = os.path.join(tempdir, '*.out')
    t = TextFilesSource(path, text_mode=True, compression=comp,
                        decoder=load)
    t.discover()
    assert t.npartitions == 2
    assert t._get_partition(0) == t.to_dask().to_delayed()[0].compute()
    out = t.read()
    assert isinstance(out, list)
    assert out[0] == data[0]


@pytest.mark.parametrize('comp', [None, 'gzip', 'bz2'])
@pytest.mark.parametrize('pars', [['msgpack.pack', 'msgpack.unpack', False],
                                  ['msgpack.packb', 'msgpack.unpackb', True],
                                  ['pickle.dump', 'pickle.load', False],
                                  ['pickle.dumps', 'pickle.loads', True]])
def test_complex_bytes(tempdir, comp, pars):
    dump, load, read = pars
    dump = import_name(dump)
    # using bytestrings means not needing extra en/decode argument to msgpack
    data = [{b'something': b'simple', b'and': 0}] * 2
    for f in ['1.out', '2.out']:
        fn = os.path.join(tempdir, f)
        with open_files([fn], mode='wb', compression=comp)[0] as fo:
            if read:
                fo.write(dump(data))
            else:
                dump(data, fo)
    # that was all setup

    path = os.path.join(tempdir, '*.out')
    t = TextFilesSource(path, text_mode=False, compression=comp,
                        decoder=load, read=read)
    t.discover()
    assert t.npartitions == 2
    assert t._get_partition(0) == t.to_dask().to_delayed()[0].compute()
    out = t.read()
    assert isinstance(out, list)
    assert out[0] == data[0]


def test_text_persist(temp_cache):
    cat = intake.open_catalog(os.path.join(here, 'sources.yaml'))
    s = cat.sometext()
    s2 = s.persist()
    assert s.read() == s2.read()


def test_text_export(temp_cache):
    import tempfile
    outdir = tempfile.mkdtemp()
    cat = intake.open_catalog(os.path.join(here, 'sources.yaml'))
    s = cat.sometext()
    out = s.export(outdir)
    fn = os.path.join(outdir, 'cat.yaml')
    with open(fn, 'w') as f:
        f.write(out.yaml())
    cat = intake.open_catalog(fn)
    s2 = cat[s.name]()
    assert s.read() == s2.read()



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os.path
import subprocess
import tempfile
import shutil

import pytest

from intake.util_tests import ex
TEST_CATALOG_YAML = os.path.join(os.path.dirname(__file__), 'catalog1.yml')


def test_list():
    cmd = [ex, '-m', 'intake.cli.client', 'list', TEST_CATALOG_YAML]
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    out, _ = process.communicate()
    out = out.decode('utf-8')

    assert len(out.strip().split('\n')) == 3
    assert "entry1" in out
    assert "entry1_part" in out
    assert "use_example1" in out


def test_full_list():
    cmd = [ex, '-m', 'intake.cli.client', 'list', '--full', TEST_CATALOG_YAML]
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    out, _ = process.communicate()
    out = out.decode('utf-8')

    assert "[entry1]" in out
    assert "[entry1_part]" in out
    assert "[use_example1]" in out


def test_describe():
    cmd = [ex, '-m', 'intake.cli.client', 'describe', TEST_CATALOG_YAML,
           'entry1']
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                               universal_newlines=True)
    out, _ = process.communicate()

    expected = """\
[entry1] args={'urlpath': '{{ CATALOG_DIR }}/entry1_*.csv'}
[entry1] container=dataframe
[entry1] description=entry1 full
[entry1] direct_access=forbid
[entry1] driver=['csv']
[entry1] metadata={'foo': 'bar', 'bar': [1, 2, 3]}
[entry1] name=entry1
[entry1] plugin=['csv']
[entry1] user_parameters=[]
"""

    assert out == expected


def test_exists_pass():
    cmd = [ex, '-m', 'intake.cli.client', 'exists', TEST_CATALOG_YAML, 'entry1']
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                               universal_newlines=True)
    out, _ = process.communicate()

    assert out == "True\n"


def test_exists_fail():
    cmd = [ex, '-m', 'intake.cli.client', 'exists', TEST_CATALOG_YAML, 'entry2']
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                               universal_newlines=True)
    out, _ = process.communicate()

    assert out == "False\n"


def test_discover():
    cmd = [ex, '-m', 'intake.cli.client', 'discover', TEST_CATALOG_YAML,
           'entry1']
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                               universal_newlines=True)
    out, _ = process.communicate()

    assert "'dtype':" in out
    assert "'metadata':" in out
    assert "'npartitions':" in out
    assert "'shape':" in out


def test_get_pass():
    cmd = [ex, '-m', 'intake.cli.client', 'get', TEST_CATALOG_YAML, 'entry1']
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                               universal_newlines=True)
    out, _ = process.communicate()

    assert 'Charlie1   25.0     3' in out
    assert 'Eve2   25.0     3' in out


def test_get_fail():
    cmd = [ex, '-m', 'intake.cli.client', 'get', TEST_CATALOG_YAML, 'entry2']
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               universal_newlines=True)
    _, err = process.communicate()

    assert "KeyError('entry2'" in err

@pytest.fixture
def temp_current_working_directory():
    prev_cwd = os.getcwd()
    dirname = tempfile.mkdtemp()
    os.chdir(dirname)

    yield dirname

    os.chdir(prev_cwd)
    shutil.rmtree(dirname)


def test_example(temp_current_working_directory):
    cmd = [ex, '-m', 'intake.cli.client', 'example']
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               universal_newlines=True)
    stdout, stderr = process.communicate()

    assert process.returncode == 0
    assert os.path.exists('us_states.yml')
    assert os.path.exists('states_1.csv')
    assert os.path.exists('states_2.csv')

    # should fail second time due to existing files
    cmd = [ex, '-m', 'intake.cli.client', 'example']
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               universal_newlines=True)
    _, err = process.communicate()

    assert process.returncode > 0



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import pytest

pn = pytest.importorskip('panel')

from ..base import Base


class BaseClass(Base):
    def __init__(self, **kwargs):
        self.panel = pn.Row()
        super().__init__(**kwargs)

    def setup(self):
        self.children = ['fake content']


def test_base_with_panel_gets_populated_when_visible_is_set_to_true():
    base = BaseClass(visible=True)
    assert base.children == ['fake content']
    assert len(base.panel.objects) == 1

    base.visible = False
    assert len(base.panel.objects) == 0
    assert base.children == ['fake content']


def test_base_with_panel_gets_populated_when_visible_is_changed_to_true():
    base = BaseClass(visible=False)
    assert len(base.panel.objects) == 0

    base.visible = True
    assert base.children == ['fake content']
    assert len(base.panel.objects) == 1




import os
import intake

catfile = os.path.join(os.path.dirname(__file__), "..", "..",
                       "catalog", "tests", "catalog_alias.yml")


def test_columns():
    cat = intake.open_catalog(catfile)
    df1 = cat.input_data.read()
    df2 = cat.derive_cols.read()
    assert df1[["state", "slug"]].equals(df2)


def _pick_columns(df, columns):
    return df[columns]


def test_df_transform():
    cat = intake.open_catalog(catfile)
    df1 = cat.input_data.read()
    df2 = cat.derive_cols_func.read()
    assert df1[["state", "slug"]].equals(df2)


def test_barebones():
    cat = intake.open_catalog(catfile)
    df1 = cat.input_data.read()
    cat = intake.open_catalog(catfile)
    assert cat.barebones.read() == len(df1)


def test_other_cat():
    cat = intake.open_catalog(catfile)
    df1 = cat.other_cat.read()
    assert df1.columns.tolist() == ["name", "score"]



"""
Test saving catalogs.
"""

import intake
from intake.catalog import Catalog
from intake.catalog.local import LocalCatalogEntry


def test_catalog_description():
    """Make sure the description comes through the save."""

    cat1 = Catalog.from_dict({
                'name': LocalCatalogEntry('name',
                                          description='description',
                                          driver=intake.catalog.local.YAMLFileCatalog,
                                            )
    },
                name='overall catalog name',
                description='overall catalog description'

    )

    cat1.save('desc_test.yaml')

    cat2 = intake.open_catalog('desc_test.yaml')

    assert cat2.description is not None



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os
import posixpath
import subprocess


def test_reset(env):
    subprocess.call(['intake', 'config', 'reset'],
                    env=env, universal_newlines=True)
    confdir = env['INTAKE_CONF_DIR']
    fn = posixpath.join(confdir, 'conf.yaml')
    assert os.path.isfile(fn)
    txt = open(fn).read()
    assert 'port: 5000' in txt


def test_info(tempdir):  # if envs is used, conf file will already exist
    env = os.environ.copy()
    env["INTAKE_CONF_DIR"] = confdir = tempdir
    out = subprocess.check_output(['intake', 'config', 'info'],
                                  env=env, universal_newlines=True)
    fn = posixpath.join(confdir, 'conf.yaml')
    assert fn in out
    assert 'INTAKE_CONF_DIR' in out
    assert 'INTAKE_CONF_FILE' not in out
    assert "(does not exist)" in out
    with open(fn, 'w') as f:
        f.write('port: 5000')
    out = subprocess.check_output(['intake', 'config', 'info'],
                                  env=env, universal_newlines=True)
    assert "(does not exist)" not in out


def test_defaults():
    out = subprocess.check_output(['intake', 'config', 'list-defaults'],
                                  universal_newlines=True)
    assert 'port: 5000' in out


def test_get(env):
    confdir = env['INTAKE_CONF_DIR']
    fn = posixpath.join(confdir, 'conf.yaml')
    with open(fn, 'w') as f:
        f.write('port: 5001')
    out = subprocess.check_output(['intake', 'config', 'get'],
                                  env=env, universal_newlines=True)
    assert 'port: 5001' in out
    out = subprocess.check_output(['intake', 'config', 'get', 'port'],
                                  env=env, universal_newlines=True)
    assert out.startswith('5001')


def test_log_level():
    env = os.environ.copy()
    env['INTAKE_LOG_LEVEL'] = 'DEBUG'
    out = subprocess.check_output(['intake', 'config', 'info'],
                                  stderr=subprocess.STDOUT,
                                  env=env, universal_newlines=True)
    assert "logger set to debug" in out



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
import pytest
pn = pytest.importorskip('panel')


@pytest.fixture
def gui(cat1, cat2):
    from ..gui import GUI
    return GUI(cats=[cat1, cat2])


def test_gui(gui, cat1, cat2, sources1):
    assert gui.cat.select.items == [cat1, cat2]
    assert gui.cats == [cat1]
    assert gui.item == sources1[0]

    assert not gui.cat.add.watchers
    assert gui.cat.add.visible is False
    assert gui.cat.add_widget.disabled is False

    assert not gui.cat.search.watchers
    assert gui.cat.search.visible is False
    assert gui.cat.search_widget.disabled is False

    assert not gui.source.plot.watchers
    assert gui.source.plot.visible is False
    assert gui.source.plot_widget.disabled is False


def test_par_selector(gui, cat2):
    gui.cat.select.selected = [cat2]
    assert gui.source.pars_widget.disabled is False

    gui.source.pars_widget.value = True
    wid = gui.source.pars_editor.panel[0]
    assert isinstance(wid, pn.widgets.Select)
    assert wid.value == ''
    s = gui.source_instance
    assert s.urlpath.endswith("crime.csv")
    wid.value = "2"
    s = gui.source_instance
    assert s.urlpath.endswith("crime2.csv")


def test_gui_remove_selected_cat(gui, cat1):
    gui.cat.select.remove_selected()
    assert gui.cat.search_widget.disabled is True
    assert gui.source.plot_widget.disabled is True
    assert gui.item is None


def test_gui_open_plot_panel(gui, cat1, cat2, sources1, sources2):
    pytest.importorskip('hvplot')
    gui.source.plot_widget.value = True
    assert gui.source.plot.visible is True
    assert len(gui.source.plot.watchers) == 2
    assert len(gui.source.plot.panel.objects) == 2
    assert gui.source.plot.source.entry == sources1[0]

    gui.source.select.cats = [cat2]
    assert gui.source.plot_widget.value is False  # hide on select change
    assert gui.source.plot.visible is False
    assert not gui.source.plot.watchers

    gui.source.plot_widget.value = True
    assert len(gui.source.plot.watchers) == 2
    assert len(gui.source.plot.panel.objects) == 2


def test_gui_open_search_panel(gui, cat1, cat2, sources1, sources2):
    gui.cat.search_widget.value = True
    assert len(gui.cat.search.watchers) == 2
    assert len(gui.cat.search.panel.objects) == 2
    assert gui.cat.search.cats == [cat1]

    gui.cat.select.selected = [cat2]
    assert len(gui.cat.search.watchers) == 2
    assert len(gui.cat.search.panel.objects) == 2
    assert gui.cat.search.cats == [cat2]

    gui.cat.search_widget.value = False
    assert not gui.cat.search.watchers
    assert gui.cat.search.visible is False


def test_gui_close_and_open_cat(gui, cat2, sources2):
    gui.cat.select.selected = [cat2]
    gui.cat.visible = False

    assert gui.cat.select.visible is False
    assert len(gui.cat.control_panel.objects) == 0
    assert gui.cat.search.visible is False
    assert gui.cat.add.visible is False

    gui.cat.visible = True
    assert gui.cat.select.visible is True
    assert len(gui.cat.control_panel.objects) == 3
    assert gui.cat.search.visible is False
    assert gui.cat.add.visible is False
    assert gui.cats == [cat2]


def test_gui_close_and_open_source(gui, cat2, sources2):
    gui.source.visible = False
    gui.cat.select.selected = [cat2]

    assert gui.source.visible is False
    assert len(gui.source.control_panel.objects) == 0
    assert gui.source.description.visible is False
    assert gui.source.plot.visible is False

    gui.source.visible = True
    assert gui.source.select.visible is True
    assert len(gui.source.control_panel.objects) == 2
    assert gui.source.description.visible is True
    assert gui.source.plot.visible is False
    assert gui.source.sources == [sources2[0]]


def test_gui_init_empty():
    from ..gui import GUI
    gui = GUI(cats=[])
    assert gui.cat.select.items == []
    assert gui.cats == []
    assert gui.item == None

    assert not gui.cat.add.watchers
    assert gui.cat.add.visible is False
    assert gui.cat.add_widget.disabled is False

    assert not gui.cat.search.watchers
    assert gui.cat.search.visible is False
    assert gui.cat.search_widget.disabled is True

    assert not gui.source.plot.watchers
    assert gui.source.plot.visible is False
    assert gui.source.plot_widget.disabled is True


def test_gui_getstate(gui, cat1, sources1):
    state = gui.__getstate__()

    assert state['visible'] is True
    assert state['cat']['visible']is True
    assert state['cat']['add']['visible'] is False
    assert state['cat']['search']['visible'] is False
    assert state['cat']['select']['selected'] == [cat1.name]
    assert state['source']['visible'] is True
    assert state['source']['select']['selected'] == [sources1[0].name]
    assert state['source']['plot']['visible'] is False


def test_gui_state_roundtrip(gui, cat1, cat2, sources1):
    from ..gui import GUI
    other = GUI.from_state(gui.__getstate__())

    assert other.cat.select.items == [cat1, cat2]
    assert other.cats == [cat1]
    assert other.sources == [sources1[0]]
    assert other.cat.search.visible is False
    assert other.cat.add.visible is False
    assert other.source.plot.visible is False



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2021, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import json
import os
from pathlib import Path
from typing import Dict, Optional

import pytest
from fsspec import open_files
from fsspec.utils import compressions

from intake.source.jsonfiles import JSONFileSource, JSONLinesFileSource

here = os.path.abspath(os.path.dirname(__file__))

EXTENSIONS = {
    compression: f".{extension}" for extension, compression in compressions.items()
}


@pytest.fixture(params=[None, "gzip", "bz2"])
def json_file(request, tmp_path) -> str:
    data = {"hello": "world"}
    file_path = str(tmp_path / "1.json")
    file_path += EXTENSIONS.get(request.param, "")
    with open_files([file_path], mode="wt", compression=request.param)[0] as f:
        f.write(json.dumps(data))
    return file_path


@pytest.fixture(params=[None, "gzip", "bz2"])
def jsonl_file(request, tmp_path) -> str:
    data = [{"hello": "world"}, [1, 2, 3]]
    file_path = str(tmp_path / "1.jsonl")
    file_path += EXTENSIONS.get(request.param, "")
    with open_files([file_path], mode="wt", compression=request.param)[0] as f:
        f.write("\n".join(json.dumps(row) for row in data))
    return file_path


def test_jsonfile(json_file: str):
    j = JSONFileSource(json_file, text_mode=True, compression="infer")
    out = j.read()
    assert isinstance(out, dict)
    assert out["hello"] == "world"


def test_jsonfile_none(json_file: str):
    try:
        j = JSONFileSource(json_file, text_mode=True, compression=None)
        out = j.read()
        # compression=None should not raise exception for uncompressed files
        assert json_file.endswith(".json")
    except UnicodeDecodeError:
        # compression=None should raise exception for compressed files
        assert not json_file.endswith(".json")
        return
    assert isinstance(out, dict)
    assert out["hello"] == "world"


def test_jsonfile_discover(json_file: str):
    j = JSONFileSource(json_file, text_mode=True, compression=None)
    schema = j.discover()
    assert schema == {'dtype': None, 'shape': None, 'npartitions': 0, 'metadata': {}}


def test_jsonlfile(jsonl_file: str):
    j = JSONLinesFileSource(jsonl_file, compression="infer")
    out = j.read()
    assert isinstance(out, list)

    assert isinstance(out[0], dict)
    assert out[0]["hello"] == "world"

    assert isinstance(out[1], list)
    assert out[1] == [1, 2, 3]


def test_jsonfilel_none(jsonl_file: str):
    try:
        j = JSONLinesFileSource(jsonl_file, compression=None)
        out = j.read()
        # compression=None should not raise exception for uncompressed files
        assert jsonl_file.endswith(".jsonl")
    except UnicodeDecodeError:
        # compression=None should raise exception for compressed files
        assert not jsonl_file.endswith(".jsonl")
        return
    assert isinstance(out, list)

    assert isinstance(out[0], dict)
    assert out[0]["hello"] == "world"

    assert isinstance(out[1], list)
    assert out[1] == [1, 2, 3]


def test_jsonfilel_discover(json_file: str):
    j = JSONLinesFileSource(jsonl_file, compression=None)
    schema = j.discover()
    assert schema == {'dtype': None, 'shape': None, 'npartitions': 0, 'metadata': {}}


def test_jsonl_head(jsonl_file: str):
    j = JSONLinesFileSource(jsonl_file, compression="infer")
    out = j.head(1)
    assert isinstance(out, list)
    assert len(out) == 1
    assert out[0]["hello"] == "world"



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os
import pytest
import requests

import intake
from intake import config
from intake.config import Config, defaults
from intake.util_tests import temp_conf, server
from intake.catalog.remote import RemoteCatalog


@pytest.mark.parametrize('conf', [
    {},
    {'port': 5000},
    {'other': True}
])
def test_load_conf(conf):
    # This test will only work if your config is set to default
    inconf = defaults.copy()
    expected = inconf.copy()
    with temp_conf(conf) as fn:
        config = Config(fn)
        config.load(fn)
        expected.update(conf)
        assert dict(config) == expected
        config.reset()
        assert dict(config) == inconf


# Tests with a real separate server process


def test_basic():
    with temp_conf({}) as fn:
        env = os.environ.copy()
        env['INTAKE_CONF_FILE'] = fn
        with server(env=env, wait=5000):
            r = requests.get('http://localhost:5000/v1/info')
            assert r.ok
    with temp_conf({}) as fn:
        env = os.environ.copy()
        env['INTAKE_CONF'] = os.path.dirname(fn)
        with server(env=env, wait=5000):
            r = requests.get('http://localhost:5000/v1/info')
            assert r.ok
    with temp_conf({}) as fn:
        env = os.environ.copy()
        env['INTAKE_CONF'] = os.path.dirname(fn) + ":/nonexistent"
        with server(env=env, wait=5000):
            r = requests.get('http://localhost:5000/v1/info')
            assert r.ok


def test_cli():
    with temp_conf({}) as fn:
        env = os.environ.copy()
        env['INTAKE_CONF_FILE'] = fn
        with server(args=('-p', '5555'), env=env, wait=5555):
            r = requests.get('http://localhost:5555/v1/info')
            assert r.ok


def test_persist_modes():
    expected_never = "never"
    expected_default = "default"

    with temp_conf({}) as fn:
        env = os.environ.copy()
        env["INTAKE_CONF_FILE"] = fn
        with server(args=("-p", "5555"), env=env, wait=5555):
            cat_never = RemoteCatalog("intake://localhost:5555", persist_mode="never")
            assert cat_never.pmode == expected_never

            cat_default = RemoteCatalog("intake://localhost:5555")
            assert cat_default.pmode == expected_default

def test_conf():
    with temp_conf({'port': 5555}) as fn:
        env = os.environ.copy()
        env['INTAKE_CONF_FILE'] = fn
        with server(env=env, wait=5555):
            r = requests.get('http://localhost:5555/v1/info')
            assert r.ok


def test_conf_auth():
    with temp_conf({'port': 5556,
                    'auth': {'cls': 'intake.auth.secret.SecretAuth',
                             'kwargs': {'secret': 'test'}}}) as fn:
        env = os.environ.copy()
        env['INTAKE_CONF_FILE'] = fn
        with server(env=env, wait=5556):
            # raw request
            r = requests.get('http://localhost:5556/v1/info')
            assert r.status_code == 403
            r = requests.get('http://localhost:5556/v1/info',
                             headers={'intake-secret': 'test'})
            assert r.ok

            # with cat
            with pytest.raises(Exception):
                intake.open_catalog('intake://localhost:5556')

            cat = intake.open_catalog('intake://localhost:5556',
                                      storage_options={'headers':
                                                       {'intake-secret': 'test'}})
            assert 'entry1' in cat


@pytest.mark.skipif(os.name == 'nt', reason="Paths are different on win")
def test_pathdirs():
    assert config.intake_path_dirs([]) == []
    assert config.intake_path_dirs(['paths']) == ['paths']
    assert config.intake_path_dirs("") == [""]
    assert config.intake_path_dirs("path1:path2") == ['path1', 'path2']
    assert config.intake_path_dirs("memory://path1:memory://path2") == [
        'memory://path1', 'memory://path2']



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------


import pytest

# module under test
import intake.cli.util as m

# TODO
def test_print_entry_info():
    pass

def test_die(capsys):
    with pytest.raises(SystemExit):
        m.die("foo")
    out, err = capsys.readouterr()
    assert err == "foo\n"
    assert out == ""

class Test_nice_join(object):

    def test_default(self):
        assert m.nice_join(["one"]) == "one"
        assert m.nice_join(["one", "two"]) == "one or two"
        assert m.nice_join(["one", "two", "three"]) == "one, two or three"
        assert m.nice_join(["one", "two", "three", "four"]) == "one, two, three or four"

    def test_string_conjunction(self):
        assert m.nice_join(["one"], conjunction="and") == "one"
        assert m.nice_join(["one", "two"], conjunction="and") == "one and two"
        assert m.nice_join(["one", "two", "three"], conjunction="and") == "one, two and three"
        assert m.nice_join(["one", "two", "three", "four"], conjunction="and") == "one, two, three and four"

    def test_None_conjunction(self):
        assert m.nice_join(["one"], conjunction=None) == "one"
        assert m.nice_join(["one", "two"], conjunction=None) == "one, two"
        assert m.nice_join(["one", "two", "three"], conjunction=None) == "one, two, three"
        assert m.nice_join(["one", "two", "three", "four"], conjunction=None) == "one, two, three, four"

    def test_sep(self):
        assert m.nice_join(["one"], sep='; ') == "one"
        assert m.nice_join(["one", "two"], sep='; ') == "one or two"
        assert m.nice_join(["one", "two", "three"], sep='; ') == "one; two or three"
        assert m.nice_join(["one", "two", "three", "four"], sep="; ") == "one; two; three or four"

class TestSubcommand(object):

    def test_initialize_abstract(self):
        with pytest.raises(NotImplementedError):
            obj = m.Subcommand("parser")
            obj.initialize()

    def test_invoke_abstract(self):
        with pytest.raises(NotImplementedError):
            obj = m.Subcommand("parser")
            obj.invoke("args")



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
import os
import posixpath
import shutil

import pytest
import yaml
from intake.utils import make_path_posix, no_duplicate_yaml


def test_windows_file_path():
    path = 'C:\\Users\\user\\fake.file'
    actual = make_path_posix(path)
    expected = 'C:/Users/user/fake.file'
    assert actual == expected


def test_make_path_posix_removes_double_sep():
    path = 'user//fake.file'
    actual = make_path_posix(path)
    expected = 'user/fake.file'
    assert actual == expected


@pytest.mark.parametrize('path', [
    '~/fake.file',
    'https://example.com',
])
def test_noops(path):
    """For non windows style paths, make_path_posix should be a noop"""
    assert make_path_posix(path) == path


def test_roundtrip_file_path():
    path = os.path.dirname(__file__)
    actual = make_path_posix(path)
    assert '\\' not in actual
    assert os.path.samefile(actual, path)


def test_yaml_tuples():
    data = (1, 2)
    text = yaml.dump(data)
    with no_duplicate_yaml():
        assert yaml.safe_load(text) == data

def copy_test_file(filename, target_dir):
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)  # can't use exist_ok in Python 2.7
    target_dir = make_path_posix(target_dir)
    # Put a catalog file in the user catalog directory
    test_dir = make_path_posix(os.path.dirname(__file__))
    test_catalog = posixpath.join(test_dir, filename)
    target_catalog = posixpath.join(target_dir, '__unit_test_'+filename)

    shutil.copyfile(test_catalog, target_catalog)
    return target_catalog



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import intake
import pytest

pytest.importorskip('panel')


def assert_widget_matches(browser):
    assert browser.options == browser.widget.options
    assert browser.selected == browser.widget.value


def test_source_browser_init_with_cats(cat1, cat2, sources1, sources2):
    from ..select import SourceSelector
    source_browser = SourceSelector(cats=[cat1, cat2])
    assert sources1[0].name in source_browser.options
    assert sources2[0].name in source_browser.options
    assert sources1[0] in source_browser.selected
    assert_widget_matches(source_browser)


def test_source_browser_set_cats(cat1, cat2, sources1, sources2):
    from ..select import SourceSelector
    source_browser = SourceSelector()
    source_browser.cats = [cat1, cat2]
    assert sources1[0].name in source_browser.options
    assert sources2[0].name in source_browser.options
    assert sources1[0] in source_browser.selected
    assert_widget_matches(source_browser)


def test_source_browser(source_browser, cat1, sources1):
    assert len(source_browser.cats) == 1
    assert cat1 in source_browser.cats
    for source in sources1:
        assert source.name in source_browser.options
    assert source_browser.selected == [sources1[0]]
    assert_widget_matches(source_browser)


def test_source_browser_add(source_browser, sources1, sources2):
    source_browser.add(sources2[0])
    for source in sources1:
        assert source.name in source_browser.options
    assert sources2[0].name in source_browser.options
    assert source_browser.selected == [sources2[0]]
    assert_widget_matches(source_browser)


def test_source_browser_add_list(source_browser, sources2):
    source_browser.add(sources2)
    assert sources2[1].name in source_browser.options
    assert source_browser.selected == [sources2[0]]
    assert_widget_matches(source_browser)


def test_source_browser_add_entry_with_nonunique_name(source_browser):
    from intake.catalog.local import LocalCatalogEntry
    name = source_browser.labels[0]
    e0 = LocalCatalogEntry(name, '', 'csv', args=dict(urlpath='foo'))
    source_browser.add(e0)
    assert f'{name}_0' in source_browser.labels
    assert name in source_browser.labels
    assert_widget_matches(source_browser)

    e1 = LocalCatalogEntry(name, '', 'csv', args=dict(urlpath='foo1'))
    source_browser.add(e1)
    assert f'{name}_1' in source_browser.labels
    assert f'{name}_0' in source_browser.labels
    assert name in source_browser.labels
    assert_widget_matches(source_browser)


def test_source_browser_remove(source_browser, sources1):
    source_browser.remove(sources1[0])
    assert sources1[0].name not in source_browser.options
    assert source_browser.selected == []
    assert_widget_matches(source_browser)


def test_source_browser_remove_list(source_browser, sources1):
    source_browser.remove(sources1)
    assert source_browser.options == {}
    assert source_browser.selected == []
    assert_widget_matches(source_browser)


def test_source_browser_select_object(source_browser, sources1):
    source_browser.selected = sources1[1]
    assert source_browser.selected == [sources1[1]]
    assert_widget_matches(source_browser)


def test_source_browser_select_name(source_browser, sources1):
    source_browser.selected = sources1[1].name
    assert source_browser.selected == [sources1[1]]
    assert_widget_matches(source_browser)


def test_source_browser_select_list_of_names(source_browser, sources1):
    source_browser.selected = []
    source_browser.selected = [source.name for source in sources1]
    assert source_browser.selected == sources1
    assert_widget_matches(source_browser)


def test_source_browser_select_list_of_objects(source_browser, sources1):
    source_browser.selected = sources1
    assert source_browser.selected == sources1
    assert_widget_matches(source_browser)



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os
import numpy as np
import posixpath
import pytest
import intake
from ..npy import NPySource

here = os.path.abspath(os.path.dirname(__file__))


@pytest.mark.parametrize('shape', [(1, ), (1, 1), (10, ), (5, 2), (3, 3, 3)])
def test_one_file(tempdir, shape):
    size = 1
    for s in shape:
        size *= s
    data = np.random.randint(1, 100, size=size).reshape(shape)
    fn = os.path.join(tempdir, 'out.npy')
    np.save(fn, data)
    s = NPySource(fn)
    out = s.read()
    assert (out == data).all()
    s = NPySource(fn, chunks=1)
    out = s.read()
    assert (out == data).all()
    s = NPySource(fn, shape=shape, dtype='int', chunks=1)
    out = s.read()
    assert (out == data).all()


@pytest.mark.parametrize('shape', [(1, ), (1, 1), (10, ), (5, 2), (3, 3, 3)])
def test_multi_file(tempdir, shape):
    size = 1
    for s in shape:
        size *= s
    data0 = np.random.randint(1, 100, size=size).reshape(shape)
    fn0 = os.path.join(tempdir, 'out0.npy')
    np.save(fn0, data0)
    data1 = np.random.randint(1, 100, size=size).reshape(shape)
    fn1 = os.path.join(tempdir, 'out1.npy')
    np.save(fn1, data1)
    data = np.stack([data0, data1])
    fn = [fn0, fn1]
    s = NPySource(fn)
    out = s.read()
    assert (out == data).all()
    s = NPySource(fn, chunks=1)
    out = s.read()
    assert (out == data).all()
    s = NPySource(fn, shape=shape, dtype='int', chunks=1)
    out = s.read()
    assert (out == data).all()
    s = NPySource(os.path.join(tempdir, 'out*.npy'))
    out = s.read()
    assert (out == data).all()


def test_zarr_minimal():
    pytest.importorskip('zarr')
    cat = intake.open_catalog(posixpath.join(here, 'sources.yaml'))
    s = cat.zarr1()
    assert s.container == 'ndarray'
    assert s.read().tolist() == [73, 98, 46, 38, 20, 12, 31, 8, 89, 72]
    assert s.npartitions == 1
    assert s.dtype == 'int'
    assert s.shape == (10,)
    assert (s.read_partition((0, )) == s.read()).all()



import copy
import glob
import os
import pytest
import sys

from ..local import YAMLFilesCatalog, MergedCatalog, EntrypointsCatalog


def test_catalog_discovery():
    basedir = os.path.dirname(__file__)
    yaml_glob = os.path.join(basedir, 'catalog_search', '*.yml')
    example_packages = os.path.join(basedir, 'catalog_search', 'example_packages')

    test_catalog = MergedCatalog([EntrypointsCatalog(paths=[example_packages]),
                                  YAMLFilesCatalog(path=[yaml_glob])])

    assert 'use_example1' in test_catalog
    assert 'ep1' in test_catalog


def test_deferred_import():
    "See https://github.com/intake/intake/pull/541"
    # We are going to mess with sys.modules here, so to be safe let's put it
    # back the way it was at the end.
    import intake.catalog
    intake.catalog.builtin = None
    mods = sys.modules.copy()
    try:
        sys.modules.pop("intake")
        sys.modules.pop("intake.catalog")
        intake.catalog.__dict__.pop('builtin')
        assert 'builtin' not in intake.catalog.__dict__
        assert intake.cat is not None
    finally:
        sys.modules.update(mods)



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
import pytest
pn = pytest.importorskip('panel')


@pytest.fixture
def description(sources1):
    from ..description import Description
    return Description(source=sources1[0])


def assert_panel_matches_contents(desc):
    """Helper function to check that panel and contents match"""
    assert desc.contents in desc.main_pane.object
    assert desc.panel.objects == [desc.main_pane]


def test_description(description):
    assert description.visible
    assert len(description.children) == 1
    assert len(description.contents) > 0
    assert_panel_matches_contents(description)


def test_description_set_source(description, sources1, sources2):
    assert description.source == sources1[0]
    description.source = sources2[0]
    assert description.source == sources2[0]
    assert_panel_matches_contents(description)


def test_description_set_source_from_list(sources2):
    from ..description import Description
    description = Description()
    description.source = sources2
    assert description.source == sources2[0]
    assert_panel_matches_contents(description)


def test_description_set_source_from_empty_list(description):
    description.source = []
    assert description.source == None
    assert "⠀" * 30 in description.contents
    assert_panel_matches_contents(description)


def test_description_clears_if_visible_is_set_to_false(description):
    description.visible = False
    assert len(description.panel.objects) == 0


def assert_is_empty(plots, visible=True):
    assert plots.source is None
    assert plots.has_plots is False
    assert plots.instructions_contents == '*No predefined plots found - declare these in the catalog*'
    assert plots.options == []
    assert plots.selected is None
    if visible:
        assert plots.instructions.object == plots.instructions_contents
        assert plots.pane.object is None
        assert len(plots.children) == 2
        assert isinstance(plots.children[-1][0][0], pn.pane.HoloViews)
        assert plots.panel.objects == plots.children
        assert len(plots.watchers) == 2
    else:
        assert not plots.selected
        assert not plots.watchers
        assert not plots.panel.objects


def assert_plotting_source2_0_line(plots, visible=True, desc=False):
    assert plots.has_plots is True
    assert plots.instructions_contents == '**Select from the predefined plots:**'
    assert plots.options == ["None", 'line_example', 'violin_example']
    if visible:
        assert plots.selected == 'None'
        plots.selected = 'line_example'
        assert plots.instructions.object == plots.instructions_contents
        assert plots.pane.object is not None
        assert len(plots.children) == 2
        assert isinstance(plots.children[-1][0][0], pn.pane.HoloViews)
        assert plots.panel.objects == plots.children
        assert len(plots.watchers) == 2
    else:
        assert not plots.selected
        assert not plots.watchers
        assert not plots.panel.objects


@pytest.fixture
def defined_plots(sources2):
    pytest.importorskip('hvplot')
    from ..defined_plots import Plots
    return Plots(source=sources2[0])


def test_defined_plots_init_empty_and_not_visible_set_source(sources2):
    pytest.importorskip('hvplot')
    from ..defined_plots import Plots
    defined_plots = Plots(source=[], visible=False)
    defined_plots.source = sources2
    assert defined_plots.source == sources2[0]()
    assert_plotting_source2_0_line(defined_plots, visible=False)


def test_defined_plots_init_with_source_not_visible_make_visible(sources2):
    pytest.importorskip('hvplot')
    from ..defined_plots import Plots
    defined_plots = Plots(source=sources2, visible=False)
    defined_plots.source = sources2
    assert defined_plots.source == sources2[0]()
    assert_plotting_source2_0_line(defined_plots, visible=False)

    defined_plots.visible = True
    assert_plotting_source2_0_line(defined_plots, visible=True)


def test_defined_plots_init_empty_and_visible():
    from ..defined_plots import Plots
    defined_plots = Plots()
    assert_is_empty(defined_plots, visible=True)


def test_defined_plots_init_empty_and_not_visible():
    from ..defined_plots import Plots
    defined_plots = Plots(visible=False)
    assert_is_empty(defined_plots, visible=False)


def test_defined_plots_set_source_to_empty_list(defined_plots):
    defined_plots.source = []
    assert_is_empty(defined_plots, visible=True)


def test_defined_plots_set_source_to_empty_list_and_visible_to_false(defined_plots):
    defined_plots.visible = False
    defined_plots.source = []
    assert_is_empty(defined_plots, visible=False)


@pytest.mark.skip(reason='This one is failing - but works in widget')
def test_defined_plots_select_a_different_plot(defined_plots):
    defined_plots.selected = 'violin_example'
    assert len(defined_plots.children) == 2
    assert isinstance(defined_plots.children[1], pn.Column)
    assert str(defined_plots.children[1].objects) == str(defined_plots.pane.objects)
    assert defined_plots.panel.objects == defined_plots.children



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os.path

import pytest
import numpy as np

from intake.container import serializer


all_serializers = pytest.mark.parametrize(
    'ser', serializer.format_registry.values())
all_compressors = pytest.mark.parametrize(
    'comp', serializer.compression_registry.values())


@all_serializers
def test_dataframe(ser):
    pd = pytest.importorskip('pandas')
    pytest.importorskip('pyarrow')
    csv_filename = os.path.join(os.path.dirname(__file__), 'entry1_1.csv')
    expected_df = pd.read_csv(csv_filename)

    # Check roundtrip
    df = ser.decode(ser.encode(expected_df, 'dataframe'), 'dataframe')

    assert expected_df.equals(df)


@all_serializers
def test_ndarray(ser):
    expected_array = np.arange(35).reshape((5, 7))

    # Check roundtrip
    array = ser.decode(ser.encode(expected_array, 'ndarray'), 'ndarray')
    np.testing.assert_array_equal(expected_array, array)


@all_serializers
def test_python(ser):
    expected = [dict(a=1, b=[1, 2], c='str'), dict(a=[1, 2], b='str', d=None)]
    actual = ser.decode(ser.encode(expected, 'python'), 'python')

    assert expected == actual


@all_compressors
def test_compression_roundtrip(comp):
    data = b'1234\x01\x02'

    assert data == comp.decompress(comp.compress(data))


def test_none_compress():
    data = b'1234\x01\x02'
    comp = serializer.NoneCompressor()

    # None should be no-op
    assert data == comp.decompress(data)
    assert data == comp.compress(data)



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os.path
import pickle

import pytest
import pandas as pd

from intake.source.tests.util import verify_datasource_interface
from .util import assert_items_equal
from intake import open_catalog
from intake.catalog.remote import RemoteCatalog

TEST_CATALOG_PATH = os.path.join(os.path.dirname(__file__), 'catalog1.yml')


def test_info_describe(intake_server):
    catalog = open_catalog(intake_server)

    assert_items_equal(list(catalog), ['use_example1', 'nested', 'entry1',
                                       'entry1_part', 'remote_env',
                                       'local_env', 'text', 'arr', 'datetime'])

    info = catalog['entry1'].describe()

    expected = {
        'container': 'dataframe',
        'description': 'entry1 full',
        'name': 'entry1',
        'direct_access': 'forbid',
        'user_parameters': []
    }
    for k in expected:
        assert info[k] == expected[k]

    info = catalog['entry1_part'].describe()

    assert info['direct_access'] == 'allow'


def test_bad_url(intake_server):
    bad_url = intake_server + '/nonsense_prefix'

    with pytest.raises(Exception):
        open_catalog(bad_url)


def test_metadata(intake_server):
    catalog = open_catalog(intake_server)
    assert hasattr(catalog, 'metadata')
    assert catalog.metadata['test'] is True
    assert catalog.version == 1


def test_nested_remote(intake_server):
    from intake.catalog.local import LocalCatalogEntry
    catalog = open_catalog()
    catalog._entries = {
        'server': LocalCatalogEntry('server', 'remote test', 'intake_remote',
                                    True, {'url': intake_server}, [],
                                    [], {}, None)
    }
    assert 'entry1' in catalog.server()


def test_remote_direct(intake_server):
    from intake.container.dataframe import RemoteDataFrame
    catalog = open_catalog(intake_server)
    s0 = catalog.entry1()
    s0.discover()
    s = RemoteDataFrame(intake_server.replace('intake', 'http'), {},
                        name='entry1', parameters={},
                        npartitions=s0.npartitions,
                        shape=s0.shape,
                        metadata=s0.metadata,
                        dtype=s0.dtype)
    assert s0.read().equals(s.read())


def test_entry_metadata(intake_server):
    catalog = open_catalog(intake_server)
    entry = catalog['arr']
    assert entry.metadata == entry().metadata


def test_unknown_source(intake_server):
    catalog = open_catalog(intake_server)

    with pytest.raises(Exception):
        catalog['does_not_exist'].describe()


def test_remote_datasource_interface(intake_server):
    catalog = open_catalog(intake_server)

    d = catalog['entry1']

    verify_datasource_interface(d)


def test_environment_evaluation(intake_server):
    catalog = open_catalog(intake_server)
    import os
    os.environ['INTAKE_TEST'] = 'client'
    catalog['remote_env']


def test_read(intake_server):
    catalog = open_catalog(intake_server)

    d = catalog['entry1']

    test_dir = os.path.dirname(__file__)
    file1 = os.path.join(test_dir, 'entry1_1.csv')
    file2 = os.path.join(test_dir, 'entry1_2.csv')
    expected_df = pd.concat((pd.read_csv(file1), pd.read_csv(file2)))
    meta = expected_df[:0]

    info = d.discover()
    assert info['dtype'] == {k: str(v) for k, v
                             in meta.dtypes.to_dict().items()}
    assert info['npartitions'] == 2
    assert info['shape'] == (None, 3)  # Do not know CSV size ahead of time

    md = d.metadata.copy()
    assert md['foo'] == 'bar'
    assert md['bar'] == [1, 2, 3]

    df = d.read()

    assert expected_df.equals(df)


def test_read_direct(intake_server):
    catalog = open_catalog(intake_server)

    d = catalog['entry1_part'].configure_new(part='2')
    test_dir = os.path.dirname(__file__)
    file2 = os.path.join(test_dir, 'entry1_2.csv')
    expected_df = pd.read_csv(file2)
    meta = expected_df[:0]

    info = d.discover()

    assert info['dtype'] == {k: str(v) for k, v
                             in meta.dtypes.to_dict().items()}
    assert info['npartitions'] == 1
    assert info['shape'] == (None, 3)  # Do not know CSV size ahead of time
    md = info['metadata'].copy()
    md.pop('catalog_dir', None)
    assert md == {'bar': [2, 4, 6], 'foo': 'baz'}

    md = d.metadata.copy()
    md.pop('catalog_dir', None)
    assert md == dict(foo='baz', bar=[2, 4, 6])
    assert d.description == 'entry1 part'
    df = d.read()

    assert expected_df.equals(df)


def test_read_chunks(intake_server):
    catalog = open_catalog(intake_server)

    d = catalog.entry1

    chunks = list(d.read_chunked())
    assert len(chunks) == 2

    test_dir = os.path.dirname(__file__)
    file1 = os.path.join(test_dir, 'entry1_1.csv')
    file2 = os.path.join(test_dir, 'entry1_2.csv')
    expected_df = pd.concat((pd.read_csv(file1), pd.read_csv(file2)))

    assert expected_df.equals(pd.concat(chunks))


def test_read_partition(intake_server):
    catalog = open_catalog(intake_server)

    d = catalog.entry1

    p2 = d.read_partition(1)
    p1 = d.read_partition(0)

    test_dir = os.path.dirname(__file__)
    file1 = os.path.join(test_dir, 'entry1_1.csv')
    file2 = os.path.join(test_dir, 'entry1_2.csv')
    assert pd.read_csv(file1).equals(p1)
    assert pd.read_csv(file2).equals(p2)


def test_close(intake_server):
    catalog = open_catalog(intake_server)

    d = catalog.entry1
    d.close()


def test_with(intake_server):
    catalog = open_catalog(intake_server)

    with catalog.entry1 as f:
        assert f.discover()


def test_pickle(intake_server):
    catalog = open_catalog(intake_server)

    d = catalog.entry1

    new_d = pickle.loads(pickle.dumps(d, pickle.HIGHEST_PROTOCOL))

    df = new_d.read()

    test_dir = os.path.dirname(__file__)
    file1 = os.path.join(test_dir, 'entry1_1.csv')
    file2 = os.path.join(test_dir, 'entry1_2.csv')
    expected_df = pd.concat((pd.read_csv(file1), pd.read_csv(file2)))

    assert expected_df.equals(df)


def test_to_dask(intake_server):
    catalog = open_catalog(intake_server)
    d = catalog.entry1
    df = d.to_dask()

    assert df.npartitions == 2


def test_remote_env(intake_server):
    import os
    os.environ['INTAKE_TEST'] = 'client'
    catalog = open_catalog(intake_server)
    catalog.remote_env
    with pytest.raises(Exception) as e:
        catalog.remote_env.read()

    with pytest.raises(Exception) as e:
        catalog.local_env
    assert 'path-client' in str(e.value)

    # prevents *client* from getting env
    catalog = open_catalog(intake_server, getenv=False)
    with pytest.raises(Exception) as e:
        catalog.local_env
    assert 'INTAKE_TEST' in str(e.value)


def test_remote_sequence(intake_server):
    import glob
    d = os.path.dirname(TEST_CATALOG_PATH)
    catalog = open_catalog(intake_server)
    assert 'text' in catalog
    s = catalog.text()
    s.discover()
    assert s.npartitions == len(glob.glob(os.path.join(d, '*.yml')))
    assert s.read_partition(0)
    assert s.read()


def test_remote_arr(intake_server):
    catalog = open_catalog(intake_server)
    assert 'arr' in catalog
    s = catalog.arr()
    s.discover()
    assert 'remote-array' in s.to_dask().name
    assert s.npartitions == 2
    assert s.read_partition(0).ravel().tolist() == list(range(50))
    assert s.read().ravel().tolist() == list(range(100))


def test_pagination(intake_server):
    PAGE_SIZE = 2
    catalog = open_catalog(intake_server, page_size=PAGE_SIZE)
    assert len(catalog._entries._page_cache) == 0
    assert len(catalog._entries._direct_lookup_cache) == 0
    # Trigger fetching one specific name.
    catalog['arr']
    assert len(catalog._entries._page_cache) == 0
    assert len(catalog._entries._direct_lookup_cache) == 1
    # Using `in` on a Catalog should not iterate.
    'arr' in catalog
    assert len(catalog._entries._page_cache) == 0
    assert len(catalog._entries._direct_lookup_cache) == 1
    # Trigger fetching just one full page.
    list(zip(range(PAGE_SIZE), catalog))
    assert len(catalog._entries._page_cache) == PAGE_SIZE
    assert len(catalog._entries._direct_lookup_cache) == 1
    # Trigger fetching all pages by list-ifying.
    list(catalog)
    assert len(catalog._entries._page_cache) > PAGE_SIZE
    assert len(catalog._entries._direct_lookup_cache) == 1
    # Now direct lookup by name should be free because everything is cached.
    catalog['text']
    assert len(catalog._entries._direct_lookup_cache) == 1


def test_dir(intake_server):
    PAGE_SIZE = 2
    catalog = open_catalog(intake_server, page_size=PAGE_SIZE)
    assert len(catalog._entries._page_cache) == 0
    assert len(catalog._entries._direct_lookup_cache) == 0
    assert not catalog._entries.complete

    with pytest.warns(UserWarning, match="Tab-complete"):
        key_completions = catalog._ipython_key_completions_()
    with pytest.warns(UserWarning, match="Tab-complete"):
        dir_ = dir(catalog)
    # __dir__ triggers loading the first page.
    assert len(catalog._entries._page_cache) == 2
    assert len(catalog._entries._direct_lookup_cache) == 0
    assert not catalog._entries.complete
    assert set(key_completions) == set(['use_example1', 'nested'])
    assert 'metadata' in dir_  # a normal attribute
    assert 'use_example1' in dir_  # an entry from the first page
    assert 'arr' not in dir_  # an entry we haven't cached yet

    # Trigger fetching one specific name.
    catalog['arr']
    with pytest.warns(UserWarning, match="Tab-complete"):
        dir_ = dir(catalog)
    with pytest.warns(UserWarning, match="Tab-complete"):
        key_completions = catalog._ipython_key_completions_()
    assert 'metadata' in dir_
    assert 'arr' in dir_  # an entry cached via direct access
    assert 'arr' in key_completions

    # Load everything.
    list(catalog)
    assert catalog._entries.complete
    with pytest.warns(None) as record:
        assert set(catalog) == set(catalog._ipython_key_completions_())
        assert set(catalog).issubset(set(dir(catalog)))
    assert len(record) == 0

    # Load without pagination (with also loads everything).
    catalog = open_catalog(intake_server, page_size=None)
    assert catalog._entries.complete
    with pytest.warns(None) as record:
        assert set(catalog) == set(catalog._ipython_key_completions_())
        assert set(catalog).issubset(set(dir(catalog)))
    assert len(record) == 0


def test_getitem_and_getattr(intake_server):
    catalog = open_catalog(intake_server)
    catalog['arr']
    with pytest.raises(KeyError):
        catalog['doesnotexist']
    with pytest.raises(KeyError):
        catalog['_doesnotexist']
    with pytest.raises(KeyError):
        # This exists as an *attribute* but not as an item.
        catalog['metadata']
    catalog.arr  # alias to catalog['arr']
    catalog.metadata  # a normal attribute
    with pytest.raises(AttributeError):
        catalog.doesnotexit
    with pytest.raises(AttributeError):
        catalog._doesnotexit
    assert catalog.arr.describe() == catalog['arr'].describe()
    assert "RemoteArray" in catalog.arr.classname
    assert isinstance(catalog.metadata, (dict, type(None)))


def test_search(intake_server):
    remote_catalog = open_catalog(intake_server)
    local_catalog = open_catalog(TEST_CATALOG_PATH)

    # Basic example
    remote_results = remote_catalog.search('entry1')
    local_results = local_catalog.search('entry1')
    expected = ['nested.entry1', 'nested.entry1_part', 'entry1', 'entry1_part']
    assert isinstance(remote_results, RemoteCatalog)
    assert list(local_results) == list(remote_results) == expected

    # Progressive search
    remote_results = remote_catalog.search('entry1').search('part')
    local_results = local_catalog.search('entry1').search('part')
    expected = ['nested.entry1_part', 'entry1_part']
    assert isinstance(remote_results, RemoteCatalog)
    assert list(local_results) == list(remote_results) == expected

    # Double progressive search
    remote_results = (remote_catalog
        .search('entry1')
        .search('part')
        .search('part'))
    local_results = (local_catalog
        .search('entry1')
        .search('part')
        .search('part'))
    expected = ['nested.entry1_part', 'entry1_part']
    assert isinstance(remote_results, RemoteCatalog)
    assert list(local_results) == list(remote_results) == expected

    # Search on a nested Catalog.
    remote_results = remote_catalog['nested'].search('entry1')
    local_results = local_catalog['nested'].search('entry1')
    expected = ['nested.entry1', 'nested.entry1_part', 'entry1', 'entry1_part']
    assert isinstance(remote_results, RemoteCatalog)
    assert list(local_results) == list(remote_results) == expected

    # Search with empty results set
    remote_results = remote_catalog.search('DOES NOT EXIST')
    local_results = local_catalog.search('DOES NOT EXIST')
    expected = []
    assert isinstance(remote_results, RemoteCatalog)
    assert list(local_results) == list(remote_results) == expected


def test_access_subcatalog(intake_server):
    catalog = open_catalog(intake_server)
    catalog['nested']


def test_len(intake_server):
    remote_catalog = open_catalog(intake_server)
    local_catalog = open_catalog(TEST_CATALOG_PATH)
    assert sum(1 for entry in local_catalog) == len(remote_catalog)


def test_datetime(intake_server):
    catalog = open_catalog(intake_server)
    info = catalog["datetime"].describe()
    catalog['datetime'].parameters['time'] == pd.Timestamp("1970")





#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import pytest
import datetime
from intake.source.utils import (
    path_to_glob, path_to_pattern, reverse_format, reverse_formats)


@pytest.mark.parametrize('pattern,expected', [
    ('data/LT05_L1TP_042033_{start_date:%Y%m%d}_{end_date:%Y%m%d}_01_T1_sr_band{band:1d}.tif',
     'data/LT05_L1TP_042033_*_*_01_T1_sr_band*.tif'),
    ('data/LT05_L1TP_042033_*_*_01_T1_sr_band*.tif',
     'data/LT05_L1TP_042033_*_*_01_T1_sr_band*.tif'),
    ('{year}/{month}/{day}.csv', '*/*/*.csv'),
    ('data/**/*.csv', 'data/**/*.csv'),
    ('data/{year:4}{month:02}{day:02}.csv', 'data/*.csv'),
    ('{lone_param}','*')
])
def test_path_to_glob(pattern, expected):
    assert path_to_glob(pattern) == expected


@pytest.mark.parametrize('pattern,resolved,expected', [
    ('*.csv', 'apple.csv', {}),
    ('{}.csv', 'apple.csv', {}),
    ('{fruit}.{}', 'apple.csv', {'fruit': 'apple'}),
    ('data//{fruit}.csv', 'data/apple.csv', {'fruit': 'apple'}),
    ('data\\{fruit}.csv', 'C:\\data\\apple.csv', {'fruit': 'apple'}),
    ('data/{fruit}.csv', 'C:\\data\\apple.csv', {'fruit': 'apple'}),
    ('data/{fruit}.csv', 'data//apple.csv', {'fruit': 'apple'}),
    ('{num:d}.csv', 'k.csv', {'num': 'k'}),
    ('{year:d}/{month:d}/{day:d}.csv', '2016/2/01.csv',
    {'year': 2016, 'month': 2, 'day': 1}),
    ('{year:.4}/{month:.2}/{day:.2}.csv', '2016/2/01.csv',
    {'year': '2016', 'month': '2', 'day': '01'}),
    ('SRLCCTabularDat/Ecoregions_{emissions}_Precip_{model}.csv',
     '/user/examples/SRLCCTabularDat/Ecoregions_a1b_Precip_ECHAM5-MPI.csv',
     {'emissions':'a1b', 'model': 'ECHAM5-MPI'}),
    ('data_{date:%Y_%m_%d}.csv', 'data_2016_10_01.csv',
     {'date': datetime.datetime(2016, 10, 1, 0, 0)}),
    ('{state}{zip:5}', 'PA19104',
    {'state': 'PA', 'zip': '19104'}),
    ('{state}{zip:5d}.csv', 'PA19104.csv',
    {'state': 'PA', 'zip': 19104}),
    ('{state:2}{zip:d}.csv', 'PA19104.csv',
    {'state': 'PA', 'zip': 19104}),
])
def test_reverse_format(pattern, resolved, expected):
    assert reverse_format(pattern, resolved) == expected


@pytest.mark.parametrize('pattern,expected', [
    ('{date:%Y%m%d}', {'date': datetime.datetime(2016, 10, 1)}),
    ('{num: .2f}', {'num': 0.23}),
    ('{percentage:.2%}', {'percentage': 0.23}),
    ('data/{year:4d}{month:02d}{day:02d}.csv',
     {'year': 2016, 'month': 10, 'day': 1})
])
def test_roundtripping_reverse_format(pattern, expected):
    resolved = pattern.format(**expected)
    actual = reverse_format(pattern, resolved)
    assert actual == expected


def test_reverse_format_errors():
    pattern = '{month}{day}{year}'
    resolved = '20012001'
    with pytest.raises(ValueError, match=("Format specifier must be set "
                                          "if no separator between fields.")):
        reverse_format(pattern, resolved)

    pattern = '{month:.2}{day:2}{year:.4}'
    resolved = '20012001'
    with pytest.raises(ValueError, match="Format specifier must have a set width"):
        reverse_format(pattern, resolved)

    pattern = '{band!s}'
    resolved = '1'
    with pytest.raises(ValueError, match="Conversion not allowed. Found on band"):
        reverse_format(pattern, resolved)

    pattern = 'data_{band}'
    resolved = '1'
    with pytest.raises(ValueError, match=("Resolved string must match "
                                            "pattern. 'data_' not found.")):

        reverse_format(pattern, resolved)


paths = ['data_2014_01_03.csv', 'data_2014_02_03.csv', 'data_2015_12_03.csv']
@pytest.mark.parametrize('pattern', [
    'data_{year}_{month}_{day}.csv',
    'data_{year:d}_{month:02d}_{day:02d}.csv',
    'data_{date:%Y_%m_%d}.csv'
])
def test_roundtrip_reverse_formats(pattern):
    args = reverse_formats(pattern, paths)
    for i, path  in enumerate(paths):
        assert pattern.format(
            **{field: values[i] for field, values in  args.items()}) == path


@pytest.mark.parametrize('path,metadata,expected', [
    ('http://data/band{band:1d}.tif',
     {'cache': [{'argkey': 'urlpath', 'regex': 'data'}]},
     '/band{band:1d}.tif'),
    ('/data/band{band:1d}.tif', {},
     '/data/band{band:1d}.tif'),
    ('/data/band{band:1d}.tif', None,
     '/data/band{band:1d}.tif')
])
def test_path_to_pattern(path, metadata, expected):
    assert path_to_pattern(path, metadata) == expected



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os
import pytest
import shutil

from intake.source.cache import FileCache, CacheMetadata
import intake
import intake.config
import logging
here = os.path.dirname(os.path.abspath(__file__))
logger = logging.getLogger('intake')


@pytest.fixture
def file_cache():
    return FileCache(
        'csv', {'argkey': 'urlpath', 'regex': 'test/path', 'type': 'file'})


def test_ensure_cache_dir(file_cache):
    file_cache._ensure_cache_dir()
    assert os.path.exists(file_cache._cache_dir)

    file_cache.clear_all()
    shutil.rmtree(file_cache._cache_dir)

    with open(file_cache._cache_dir, 'w') as f:
        f.write('')

    with pytest.raises(Exception):
        file_cache._ensure_cache_dir()

    os.remove(file_cache._cache_dir)

    file_cache.clear_all()


def test_munge_path(file_cache):
    subdir = 'subdir'
    cache_path = file_cache._munge_path(subdir, 'test/path/foo.cvs')
    assert subdir in cache_path
    assert 'test/path' not in cache_path

    file_cache._spec['regex'] = 'https://example.com'
    cache_path = file_cache._munge_path(subdir, 'https://example.com/catalog.yml')
    assert subdir in cache_path
    assert file_cache._cache_dir in cache_path
    assert 'http' not in cache_path


def test_hash(file_cache):
    subdir = file_cache._hash('foo/bar.csv')

    import string
    # Checking for md5 hash
    assert all(c in string.hexdigits for c in subdir)

    file_cache._driver = 'bar'
    subdir_new = file_cache._hash('foo/bar.csv')
    assert subdir_new != subdir

    file_cache._driver = 'csv'
    subdir_new = file_cache._hash('foo/bar.csv')
    assert subdir_new == subdir

    file_cache._spec['regex'] = 'foo/bar'
    subdir_new = file_cache._hash('foo/bar.csv')
    assert subdir_new != subdir


def test_path(file_cache):
    urlpath = 'test/path/foo.csv'
    file_cache._spec['regex'] = 'test/path/'
    cache_path = file_cache._path(urlpath)

    assert file_cache._cache_dir in cache_path
    assert '//' not in cache_path[1:]
    file_cache.clear_all()


def test_dir_cache(tempdir, temp_cache):
    [os.makedirs(os.path.join(tempdir, d)) for d in [
        'main', 'main/sub1', 'main/sub2']]
    for f in ['main/afile', 'main/sub1/subfile', 'main/sub2/subfile1',
              'main/sub2/subfile2']:
        fn = os.path.join(tempdir, f)
        with open(fn, 'w') as fo:
            fo.write(f)
    fn = os.path.join(tempdir, 'cached.yaml')
    shutil.copy2(os.path.join(here, 'cached.yaml'), fn)
    cat = intake.open_catalog(fn)
    s = cat.dirs()
    out = s.cache[0].load(s._urlpath, output=False)
    assert out[0] == os.path.join(tempdir, s.cache[0]._path(s._urlpath))
    assert open(os.path.join(out[0], 'afile')).read() == 'main/afile'
    md = CacheMetadata()
    got = md[s._urlpath]

    # Avoid re-copy
    s = cat.dirs()
    s.cache[0].load(s._urlpath, output=False)
    md2 = CacheMetadata()
    got2 = md2[s._urlpath]
    assert got == got2


def test_compressed_cache(temp_cache):
    cat = intake.open_catalog(os.path.join(here, 'cached.yaml'))
    s = cat.calvert()
    old = intake.config.conf['cache_download_progress']
    try:
        intake.config.conf['cache_download_progress'] = False
        df = s.read()
        assert len(df)
        md = CacheMetadata()
        assert len(md[s._urlpath]) == 1  # we gained exactly one CSV
        intake.config.conf['cache_download_progress'] = False
        df = s.read()
        assert len(df)
        md = CacheMetadata()
        assert len(md[s._urlpath]) == 1  # we still have exactly one CSV
    finally:
        intake.config.conf['cache_download_progress'] = old


def test_filtered_compressed_cache(temp_cache):
    cat = intake.open_catalog(os.path.join(here, 'cached.yaml'))
    s = cat.calvert_filter()
    old, intake.config.conf['cache_download_progress'] = intake.config.conf['cache_download_progress'], False
    try:
        df = s.read()
        assert len(df)
        md = CacheMetadata()
        assert len(md[s._urlpath]) == 1  # we gained exactly one CSV
        intake.config.conf['cache_download_progress'] = False
        df = s.read()
        assert len(df)
        md = CacheMetadata()
        assert len(md[s._urlpath]) == 1  # we still have exactly one CSV
    finally:
        intake.config.conf['cache_download_progress'] = old


def test_cache_to_cat(tempdir):
    old = intake.config.conf.copy()
    olddir = intake.config.confdir
    intake.config.confdir = str(tempdir)
    intake.config.conf.update({'cache_dir': 'catdir',
                               'cache_download_progress': False,
                               'cache_disabled': False})
    try:
        fn0 = os.path.join(here, 'calvert_uk.zip')
        fn1 = os.path.join(tempdir, 'calvert_uk.zip')
        shutil.copy2(fn0, fn1)
        fn0 = os.path.join(here, 'cached.yaml')
        fn1 = os.path.join(tempdir, 'cached.yaml')
        shutil.copy2(fn0, fn1)
        cat = intake.open_catalog(fn1)
        s = cat.calvert()
        df = s.read()
        assert len(df)
        md = CacheMetadata()
        f = md[s._urlpath][0]
        assert f['cache_path'].startswith(str(tempdir))
        assert 'intake_cache' in os.listdir(tempdir)
        assert os.listdir(os.path.join(tempdir, 'intake_cache'))
    finally:
        intake.config.confdir = olddir
        intake.config.conf.update(old)


def test_compressed_cache_infer(temp_cache):
    cat = intake.open_catalog(os.path.join(here, 'cached.yaml'))
    s = cat.calvert_infer()
    old = intake.config.conf['cache_download_progress']
    try:
        intake.config.conf['cache_download_progress'] = False
        df = s.read()
        assert len(df)
    finally:
        intake.config.conf['cache_download_progress'] = old


@pytest.mark.skipif(os.name == 'nt', reason="No CLI tools on windows")
@pytest.mark.parametrize('comp', ['tgz', 'tbz', 'tar', 'gz', 'bz'])
def test_compressions(temp_cache, tempdir, comp):
    from intake.source.cache import CompressedCache
    import shlex
    import subprocess
    data = b'hello'
    fn = os.path.join(tempdir, 'data')
    with open(fn, 'wb') as f:
        f.write(data)
    if comp == 'tgz':
        subprocess.call(shlex.split('tar -czf {fn}.tgz {fn} -C {d}'.format(
            fn=fn, d=tempdir)))
    elif comp == 'tbz':
        subprocess.call(shlex.split('tar -cjf {fn}.tbz {fn} -C {d}'.format(
            fn=fn, d=tempdir)))
    elif comp == 'tar':
        subprocess.call(shlex.split('tar -cf {fn}.tar {fn} -C {d}'.format(
            fn=fn, d=tempdir)))
    elif comp == 'gz':
        subprocess.call(shlex.split('gzip ' + fn))
    elif comp == 'bz':
        subprocess.call(shlex.split('bzip2 ' + fn))
    fn = os.path.join(
        tempdir,
        [f for f in os.listdir(tempdir) if f.startswith('data.')][0])
    comp = CompressedCache(driver=None, spec={'decomp': comp})
    comp.output = None
    files_in, files_out = comp._make_files(fn)
    out = comp._load(files_in, files_out, fn, meta=False)
    with open(out[0], 'rb') as f:
        assert f.read() == b'hello'

    comp = CompressedCache(driver=None, spec={'decomp': 'infer'})
    comp.output = None
    files_in, files_out = comp._make_files(fn)
    out = comp._load(files_in, files_out, fn, meta=False)
    with open(out[0], 'rb') as f:
        assert f.read() == b'hello'


def test_compressed_cache_bad(temp_cache):
    cat = intake.open_catalog(os.path.join(here, 'cached.yaml'))
    s = cat.calvert_badkey()
    old = intake.config.conf['cache_download_progress']
    try:
        intake.config.conf['cache_download_progress'] = False
        with pytest.raises(ValueError):
            s.read()
    finally:
        intake.config.conf['cache_download_progress'] = old


@pytest.mark.xfail
def test_dat(temp_cache):
    import subprocess
    import time
    try:
        subprocess.call(['dat', '-v'])
    except Exception:
        pytest.skip("DAT not avaiable")
    P = subprocess.Popen(['dat', 'share', '-d',
                          os.path.join(here, 'test_dat/')])
    time.sleep(1.5)
    try:
        cat = intake.open_catalog(os.path.join(here, 'cached.yaml'))
        out = cat.dat_data.read()
        assert out[0]['title'] == 'intake test copy'
    finally:
        P.terminate()



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from __future__ import print_function

import os

install_local_data_files =  "{{cookiecutter.install_local_data_files}}" == "yes"

print("Don't forget to edit {{cookiecutter.package_name}}/{{cookiecutter.dataset_name}}.yaml to add your data sources!")

if install_local_data_files:
    os.mkdir('src')
    print("Put your data files in the {{cookiecutter.package_name}}/src/ directory to be included in the package.")



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from collections import defaultdict
import pickle

import pytest
import numpy as np
import pandas as pd

import intake.source.base as base
import intake.source.derived as der
import intake.source


def test_datasource_base_method_exceptions():
    # Unimplemented methods should raise exceptions
    d = base.DataSource()

    for (method_name, args) in [('_get_schema', []),
                                ('_get_partition', [1]),
                                ('_close', [])]:
        method = getattr(d, method_name)
        with pytest.raises(NotImplementedError):
            method(*args)


def test_name():
    d = base.DataSource()
    assert d.classname == 'intake.source.base.DataSource'
    assert isinstance(hash(d), int)


def test_datasource_base_context_manager():
    # Base data source should raise a "need to implement" exception when it
    # enters the context manager (which loads the schema)

    with pytest.raises(NotImplementedError):
        with base.DataSource():
            pass


class MockDataSourceDataFrame(base.DataSource):
    """Mock Data Source subclass that returns dataframe containers

    Used to verify that the base DataSource class logic works for dataframes.
    """

    container = 'dataframe'
    name = 'mock'

    def __init__(self, a, b):
        self.a = a
        self.b = b

        self.call_count = defaultdict(lambda: 0)

        super(MockDataSourceDataFrame, self).__init__(
            metadata=dict(a=1, b=2)
        )
        self.npartitions = 2

    def _get_schema(self):
        self.call_count['_get_schema'] += 1

        return base.Schema(dtype=np.dtype([('x', np.int64), ('y', np.int64)]),
                           shape=(6,),
                           npartitions=2,
                           extra_metadata=dict(c=3, d=4))

    def _get_partition(self, i):
        self.call_count['_get_partition'] += 1

        if i == 0:
            return pd.DataFrame({'x': [1, 2, 3], 'y': [10, 20, 30]})
        elif i == 1:
            return pd.DataFrame({'x': [4, 5, 6], 'y': [40, 50, 60]})
        else:
            raise Exception('This should never happen')

    def read(self):
        return pd.concat([self._get_partition(i)
                          for i in range(self.npartitions)])

    def to_dask(self):
        import dask.dataframe as dd
        import dask
        return dd.from_delayed([dask.delayed(self._get_partition)(i)
                                for i in range(self.npartitions)])

    def _close(self):
        self.call_count['_close'] += 1


@pytest.fixture
def source_dataframe():
    return MockDataSourceDataFrame(a=1, b=2)


def test_datasource_discover(source_dataframe):
    r = source_dataframe.discover()

    assert source_dataframe.container == 'dataframe'

    row_dtype = np.dtype([('x', np.int64), ('y', np.int64)])
    assert r == {
        'dtype': row_dtype,
        'shape': (6,),
        'npartitions': 2,
        'metadata': dict(a=1, b=2, c=3, d=4),
    }

    # check attributes have been set
    assert source_dataframe.dtype == row_dtype
    assert source_dataframe.shape == (6,)
    assert source_dataframe.npartitions == 2
    assert source_dataframe.metadata == dict(a=1, b=2, c=3, d=4)

    # check that _get_schema is only called once
    assert source_dataframe.call_count['_get_schema'] == 1
    source_dataframe.discover()
    assert source_dataframe.call_count['_get_schema'] == 1


def check_df(data):
    pd.testing.assert_frame_equal(
        data.reset_index(drop=True),
        pd.DataFrame({
            'x': [1, 2, 3, 4, 5, 6],
            'y': [10, 20, 30, 40, 50, 60]}),
        check_index_type=False,
    )


def test_datasource_read(source_dataframe):
    data = source_dataframe.read()

    check_df(data)


def check_df_parts(parts):
    assert len(parts) == 2
    pd.testing.assert_frame_equal(
        parts[0],
        pd.DataFrame({
            'x': [1, 2, 3],
            'y': [10, 20, 30]})
    )
    pd.testing.assert_frame_equal(
        parts[1],
        pd.DataFrame({
            'x': [4, 5, 6],
            'y': [40, 50, 60]})
    )


def test_datasource_read_chunked(source_dataframe):
    parts = [p for p in source_dataframe.read_chunked()]

    check_df_parts(parts)


def test_datasource_read_partition(source_dataframe):
    source_dataframe.discover()

    parts = [source_dataframe.read_partition(i)
             for i in range(source_dataframe.npartitions)]

    check_df_parts(parts)


def test_datasource_read_partition_out_of_range(source_dataframe):
    with pytest.raises(IndexError):
        source_dataframe.read_partition(-1)

    with pytest.raises(IndexError):
        source_dataframe.read_partition(2)


def test_datasource_to_dask(source_dataframe):
    ddf = source_dataframe.to_dask()

    check_df(ddf.compute())


def test_datasource_close(source_dataframe):
    source_dataframe.close()

    assert source_dataframe.call_count['_close'] == 1


def test_datasource_context_manager(source_dataframe):
    with source_dataframe:
        pass

    assert source_dataframe.call_count['_close'] == 1


def test_datasource_pickle(source_dataframe):
    pkl = pickle.dumps(source_dataframe)
    new_obj = pickle.loads(pkl)

    check_df(new_obj.read())


class MockDataSourcePython(base.DataSource):
    """Mock Data Source subclass that returns Python list containers

    Used to verify that the base DataSource class logic works for Python lists.
    """
    container = 'python'
    name = 'mock'

    def __init__(self, a, b):
        self.a = a
        self.b = b

        self.call_count = defaultdict(lambda: 0)

        super(MockDataSourcePython, self).__init__(
            metadata=dict(a=1, b=2)
        )
        self.npartitions = 2

    def _get_schema(self):
        self.call_count['_get_schema'] += 1

        return base.Schema(dtype=None,
                           shape=(4,),
                           npartitions=2,
                           extra_metadata=dict(c=3, d=4))

    def _get_partition(self, i):
        self.call_count['_get_partition'] += 1

        if i == 0:
            return [{'x': 'foo', 'y': 'bar'},
                    {'x': 'foo', 'y': 'bar', 'z': 'baz'}]
        elif i == 1:
            return [{'x': 1}, {}]
        else:
            raise Exception('This should never happen')

    def read(self):
        return sum([self._get_partition(i) for i in range(self.npartitions)],
               [])

    def to_dask(self):
        import dask.bag as db
        import dask
        return db.from_delayed([dask.delayed(self._get_partition)(i)
                                for i in range(self.npartitions)])

    def _close(self):
        self.call_count['_close'] += 1


@pytest.fixture
def source_python():
    return MockDataSourcePython(a=1, b=2)


def test_datasource_python_discover(source_python):
    r = source_python.discover()

    assert source_python.container == 'python'

    assert r == {
        'dtype': None,
        'shape': (4,),
        'npartitions': 2,
        'metadata': dict(a=1, b=2, c=3, d=4),
    }

    # check attributes have been set
    assert source_python.dtype is None
    assert source_python.shape == (4,)
    assert source_python.npartitions == 2
    assert source_python.metadata == dict(a=1, b=2, c=3, d=4)

    # check that _get_schema is only called once
    assert source_python.call_count['_get_schema'] == 1
    source_python.discover()
    assert source_python.call_count['_get_schema'] == 1


def test_datasource_python_read(source_python):
    data = source_python.read()

    assert data == [{'x': 'foo', 'y': 'bar'},
                    {'x': 'foo', 'y': 'bar', 'z': 'baz'},
                    {'x': 1}, {}]


def test_datasource_python_to_dask(source_python):
    db = list(source_python.to_dask())

    assert db == [{'x': 'foo', 'y': 'bar'},
                  {'x': 'foo', 'y': 'bar', 'z': 'baz'},
                  {'x': 1}, {}]


def test_yaml_method(source_python):
    out = source_python.yaml()
    assert 'mock' in out  # the "driver"
    assert 'metadata' in out
    assert 'a: 1' in out


def test_alias_fail():
    s = der.AliasSource('atarget')
    s.container == 'other'
    with pytest.raises(ValueError):
        s.read()


def test_reconfigure():
    s = MockDataSourcePython(a=1, b=2)
    assert s.a == 1
    s2 = s(a=3, b=4)
    assert s2.a == 3
    s3 = s2(a=4)
    assert s3.a == 4
    assert s3.b == 4


@pytest.mark.parametrize("data", [
    ("intake.source.import_name", intake.source.import_name),
    ("intake.source:import_name", intake.source.import_name),
    ("intake.source.DataSource", intake.source.DataSource),
    ("intake.source:DataSource", intake.source.DataSource),
    ("intake.source:base.DataSource", intake.source.DataSource),
])
def test_import_name(data):
    text, object = data
    assert intake.source.import_name(text) == object



import pytest
from intake.catalog.base import Catalog


def test_no_entry():
    cat = Catalog()
    cat2 = cat.configure_new()
    assert isinstance(cat2, Catalog)
    assert cat.auth is None
    assert cat2.auth is None


def test_regression():
    with pytest.raises(ValueError):
        Catalog("URI")



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os
import os.path
import shutil
import tempfile
import time

import pytest

from .util import assert_items_equal
from intake import open_catalog

TMP_DIR = tempfile.mkdtemp()
TEST_CATALOG_PATH = [TMP_DIR]

YAML_FILENAME = 'intake_test_catalog.yml'
MISSING_PATH = os.path.join(TMP_DIR, 'a')


def teardown_module(module):
    try:
        shutil.rmtree(TMP_DIR)
    except:
        pass


@pytest.fixture
def intake_server_with_config(intake_server):
    fullname = os.path.join(TMP_DIR, YAML_FILENAME)

    with open(fullname, 'w') as f:
        f.write('''
plugins:
  source:
    - module: intake.catalog.tests.example1_source
    - module: intake.catalog.tests.example_plugin_dir.example2_source
sources:
  use_example1:
    description: example1 source plugin
    driver: example1
    args: {}
''')

    time.sleep(1)

    yield intake_server
    os.remove(fullname)


def test_reload_updated_config(intake_server_with_config):
    catalog = open_catalog(intake_server_with_config, ttl=0.1)

    entries = list(catalog)
    assert entries == ['use_example1']

    with open(os.path.join(TMP_DIR, YAML_FILENAME), 'a') as f:
        f.write('''
  use_example1_1:
    description: example1 other
    driver: example1
    args: {}
''')

    time.sleep(1.2)

    assert_items_equal(list(catalog), ['use_example1', 'use_example1_1'])


def test_reload_updated_directory(intake_server_with_config):
    catalog = open_catalog(intake_server_with_config, ttl=0.1)

    orig_entries = list(catalog)
    assert 'example2' not in orig_entries

    filename = os.path.join(TMP_DIR, 'intake_test_catalog2.yml')
    with open(filename, 'w') as f:
        f.write('''
sources:
  example2:
    description: source 2
    driver: csv
    args:
        urlpath: none
        ''')

    time.sleep(1.2)

    assert_items_equal(list(catalog), ['example2'] + orig_entries)


def test_reload_missing_remote_directory(intake_server):
    try:
        shutil.rmtree(TMP_DIR)
    except:
        pass

    time.sleep(1)
    catalog = open_catalog(intake_server, ttl=0.1)
    assert_items_equal(list(catalog), [])

    os.mkdir(TMP_DIR)
    with open(os.path.join(TMP_DIR, YAML_FILENAME), 'w') as f:
        f.write('''
plugins:
  source:
    - module: intake.catalog.tests.example1_source
    - module: intake.catalog.tests.example_plugin_dir.example2_source
sources:
  use_example1:
    description: example1 source plugin
    driver: example1
    args: {}
        ''')
    time.sleep(1.2)

    assert_items_equal(list(catalog), ['use_example1'])
    try:
        shutil.rmtree(TMP_DIR)
    except:
        pass


def test_reload_missing_local_directory(tempdir):
    catalog = open_catalog(tempdir + '/*', ttl=0.1)
    assert_items_equal(list(catalog), [])

    with open(os.path.join(tempdir, YAML_FILENAME), 'w') as f:
        f.write('''
plugins:
  source:
    - module: intake.catalog.tests.example1_source
    - module: intake.catalog.tests.example_plugin_dir.example2_source
sources:
  use_example1:
    description: example1 source plugin
    driver: example1
    args: {}
        ''')

    time.sleep(1.2)
    assert 'use_example1' in catalog



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
from distutils.version import LooseVersion
import pytest
pn = pytest.importorskip('panel')
too_old = LooseVersion(pn.__version__) < LooseVersion("0.9.5")


@pytest.mark.skipif(too_old, reason="Use with latest panel")
@pytest.fixture
def gui(sources1):
    from ..gui import SourceGUI
    return SourceGUI(sources=sources1)


def test_gui_attribute(sources1):
    assert sources1[0].gui


def test_gui(gui, sources1):
    assert gui.select.items == sources1
    assert gui.sources == [sources1[0]]

    assert not gui.plot.watchers
    assert gui.plot.visible is False
    assert gui.plot_widget.disabled is False


def test_gui_close_and_open_select(gui, sources1):
    gui.select.selected = [sources1[1]]
    gui.select.visible = False
    assert not gui.select.watchers

    gui.select.visible = True
    assert len(gui.select.watchers) == 1
    assert gui.select.selected == [sources1[1]]
    assert gui.plot_widget.disabled is False


def test_gui_getstate(gui, sources1):
    state = gui.__getstate__()

    assert state['visible']is True
    assert state['plot']['visible'] is False
    assert state['select']['selected'] == [sources1[0].name]


def test_gui_state_roundtrip(gui, sources1):
    from ..gui import SourceGUI
    other = SourceGUI.from_state(gui.__getstate__())

    assert other.select.items == sources1
    assert other.sources == [sources1[0]]
    assert other.plot.visible is False
    assert other.description.visible is True


def test_gui_state_roundtrip_with_subpanels(gui, sources1):
    from ..gui import SourceGUI
    gui.plot.visible = True

    other = SourceGUI.from_state(gui.__getstate__())

    assert other.select.items == sources1
    assert other.sources == [sources1[0]]
    assert other.plot.visible is True
    assert other.plot_widget.value is True



#!/usr/bin/env python
#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os
import posixpath
import pytest
from intake.container.dataframe import GenericDataFrame
here = os.path.abspath(os.path.dirname(__file__))


def test_generic_dataframe():
    pd = pytest.importorskip('pandas')

    def make_a_part(openfile):
        return pd.DataFrame([[0]], columns=['x'])

    url = posixpath.join(here, '*.py')
    s = GenericDataFrame(url, reader=make_a_part)
    ddf = s.to_dask()
    assert ddf.compute().x.unique() == [0]
    df = s.read()
    assert len(df) == len(ddf)





#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from intake.source import base, registry


def assert_items_equal(a, b):
    assert len(a) == len(b) and sorted(a) == sorted(b)


class TestingSource(base.DataSource):
    """A source that gives back whatever parameters were passed to it"""
    name = 'test'
    version = '0.0.1'
    container = 'python'
    partition_access = False

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        super(TestingSource, self).__init__('python')
        self.npartitions = 1

    def _load_metadata(self):
        pass

    def _get_partition(self, _):
        return self.args, self.kwargs


def register():
    registry['test'] = TestingSource



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os.path
import posixpath
import pickle

import pytest
import pandas as pd

import intake.source.csv as csv
from intake.utils import make_path_posix
from .util import verify_plugin_interface, verify_datasource_interface


@pytest.fixture
def data_filenames():
    basedir = make_path_posix(os.path.dirname(__file__))
    return dict(sample1=posixpath.join(basedir, 'sample1.csv'),
                sample2_1=posixpath.join(basedir, 'sample2_1.csv'),
                sample2_2=posixpath.join(basedir, 'sample2_2.csv'),
                sample2_all=posixpath.join(basedir, 'sample2_*.csv'),
                sample_pattern=posixpath.join(basedir, 'sample{num:d}_{dup:d}.csv'))


@pytest.fixture
def sample1_datasource(data_filenames):
    return csv.CSVSource(data_filenames['sample1'])


@pytest.fixture
def sample2_datasource(data_filenames):
    return csv.CSVSource(data_filenames['sample2_all'])


@pytest.fixture
def sample_pattern_datasource(data_filenames):
    return csv.CSVSource(data_filenames['sample_pattern'])

@pytest.fixture
def sample_list_datasource(data_filenames):
    return csv.CSVSource([data_filenames['sample2_1'], data_filenames['sample2_2']])

@pytest.fixture
def sample_list_datasource_with_path_as_pattern_str(data_filenames):
    return csv.CSVSource(
        [data_filenames['sample2_1'], data_filenames['sample2_2']],
        path_as_pattern='sample{num:d}_{dup:d}.csv')

@pytest.fixture
def sample_pattern_datasource_with_cache(data_filenames):
    metadata = {'cache': [{'argkey': 'urlpath',
                           'regex': make_path_posix(
                                os.path.dirname(__file__)),
                           'type': 'file'}]}
    return csv.CSVSource(data_filenames['sample_pattern'], metadata=metadata)


def test_csv_plugin():
    p = csv.CSVSource
    assert isinstance(p.version, str)
    assert p.container == 'dataframe'
    verify_plugin_interface(p)


def test_open(data_filenames):
    d = csv.CSVSource(data_filenames['sample1'])
    assert d.container == 'dataframe'
    assert d.description is None
    verify_datasource_interface(d)


def test_discover(sample1_datasource):
    info = sample1_datasource.discover()

    assert info['dtype'] == {'name': 'object', 'score': 'float64',
                             'rank': 'int64'}
    # Do not know length without parsing CSV
    assert info['shape'] == (None, 3)
    assert info['npartitions'] == 1


def test_read(sample1_datasource, data_filenames):
    expected_df = pd.read_csv(data_filenames['sample1'])
    df = sample1_datasource.read()

    assert expected_df.equals(df)

def test_read_list(sample_list_datasource,  data_filenames):
    df_1 = pd.read_csv(data_filenames['sample2_1'])
    df_2 = pd.read_csv(data_filenames['sample2_2'])
    expected_df = pd.concat([df_1, df_2])
    df = sample_list_datasource.read()

    assert expected_df.equals(df)

def test_read_chunked(sample1_datasource, data_filenames):
    expected_df = pd.read_csv(data_filenames['sample1'])

    parts = list(sample1_datasource.read_chunked())
    df = pd.concat(parts)

    assert expected_df.equals(df)


def check_read_pattern_output(source):
    da = source.to_dask()
    assert da.num.cat.known is True
    assert da.dup.cat.known is True

    # check that first partition has correct num and dup; which file
    # it represents is not guaranteed
    df0 = da.get_partition(0).compute()
    if df0['name'][0].endswith('1'):
        assert all(df0.num == 2)
        assert all(df0.dup == 1)
    elif df0['name'][0].endswith('2'):
        assert all(df0.num == 2)
        assert all(df0.dup == 2)
    elif df0['name'][0].endswith('3'):
        assert all(df0.num == 3)
        assert all(df0.dup == 2)

    df = source.read()
    assert len(df.columns) == 5
    assert 'num' in df
    assert 'dup' in df
    assert df.num.dtype == 'category'
    assert df.dup.dtype == 'category'

    names = ['Alice', 'Bob', 'Charlie', 'Eve']

    file_1 = df[df['name'].isin(['{}1'.format(name) for name in names])]
    assert all(file_1.num == 2)
    assert all(file_1.dup == 1)

    file_2 = df[df['name'].isin(['{}2'.format(name) for name in names])]
    assert all(file_2.num == 2)
    assert all(file_2.dup == 2)

    file_3 = df[df['name'].isin(['{}3'.format(name) for name in names])]
    assert all(file_3.num == 3)
    assert all(file_3.dup == 2)

    return da


def test_read_pattern(sample_pattern_datasource):
    da = sample_pattern_datasource.to_dask()
    assert set(da.num.cat.categories) == {2, 3}
    assert set(da.dup.cat.categories) == {1, 2}
    check_read_pattern_output(sample_pattern_datasource)


def test_read_pattern_with_cache(sample_pattern_datasource_with_cache):
    da = sample_pattern_datasource_with_cache.to_dask()
    assert set(da.num.cat.categories) == {2, 3}
    assert set(da.dup.cat.categories) == {1, 2}
    check_read_pattern_output(sample_pattern_datasource_with_cache)


def test_read_pattern_with_path_as_pattern_str(sample_list_datasource_with_path_as_pattern_str):
    da = sample_list_datasource_with_path_as_pattern_str.to_dask()
    assert set(da.num.cat.categories) == {2}
    assert set(da.dup.cat.categories) == {1, 2}
    check_read_pattern_output(sample_list_datasource_with_path_as_pattern_str)


def test_read_partition(sample2_datasource, data_filenames):
    expected_df1 = pd.read_csv(data_filenames['sample2_1'])
    expected_df2 = pd.read_csv(data_filenames['sample2_2'])

    sample2_datasource.discover()
    assert sample2_datasource.npartitions == 2

    # Read partitions is opposite order
    df2 = sample2_datasource.read_partition(1)
    df1 = sample2_datasource.read_partition(0)

    assert expected_df1.equals(df1)
    assert expected_df2.equals(df2)


def test_to_dask(sample1_datasource, data_filenames):
    dd = sample1_datasource.to_dask()
    df = dd.compute()

    expected_df = pd.read_csv(data_filenames['sample1'])

    assert expected_df.equals(df)


def test_plot(sample1_datasource):
    pytest.importorskip('hvplot')
    import holoviews

    p = sample1_datasource.plot()
    assert isinstance(p, holoviews.NdOverlay)


def test_close(sample1_datasource, data_filenames):
    sample1_datasource.close()
    # Can reopen after close
    df = sample1_datasource.read()
    expected_df = pd.read_csv(data_filenames['sample1'])

    assert expected_df.equals(df)


def test_pickle(sample1_datasource):
    pickled_source = pickle.dumps(sample1_datasource)
    sample1_clone = pickle.loads(pickled_source)

    expected_df = sample1_datasource.read()
    df = sample1_clone.read()

    assert expected_df.equals(df)



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os
import pytest
import shutil
import string
import time

import intake
from intake.config import conf
from intake.utils import make_path_posix


@pytest.fixture
def catalog_cache():
    path = os.path.dirname(__file__)
    return intake.open_catalog(os.path.join(path, 'catalog_caching.yml'))


def test_load_csv(catalog_cache, tempdir):
    os.environ['TEST_CACHE_DIR'] = str(tempdir)

    catalog_cache['test_cache_new'].read()
    files = os.listdir(tempdir)
    assert 'cache' in files
    assert len(files) == 2
    cache_id = [f for f in files if f != 'cache'][0]
    assert all(c in string.hexdigits for c in cache_id)


def test_list_of_files(catalog_cache):
    pd = pytest.importorskip('pandas')
    s1 = catalog_cache['test_cache']
    s2 = catalog_cache['test_list_cache']
    assert s2.read().equals(pd.concat([s1.read(), s1.read()]))


def test_bad_type_cache(catalog_cache):
    with pytest.raises(IndexError):
        catalog_cache['test_bad_type_cache_spec'].cache


def test_load_textfile(catalog_cache):
    cat = catalog_cache['text_cache']
    cache = cat.cache[0]

    cache_paths = cache.load(cat._urlpath, output=False)
    cache_path = cache_paths[-1]

    assert cache._cache_dir in cache_path
    assert os.path.isfile(cache_path)

    cache_id = os.path.basename(os.path.dirname(cache_path))
    # Checking for md5 hash
    assert all(c in string.hexdigits for c in cache_id)
    cache.clear_all()


def test_load_arr(catalog_cache):
    cat = catalog_cache['arr_cache']
    cache = cat.cache[0]

    cache_paths = cache.load(cat.path, output=False)
    cache_path = cache_paths[-1]

    assert cache._cache_dir in cache_path
    assert os.path.isfile(cache_path)

    cache_id = os.path.basename(os.path.dirname(cache_path))
    # Checking for md5 hash
    assert all(c in string.hexdigits for c in cache_id)
    cache.clear_all()


@pytest.mark.parametrize('section', ['test_no_regex',
                                     'test_regex_no_match',
                                     'test_regex_partial_match'])
def test_regex(catalog_cache, section):
    cat = catalog_cache[section]
    cache = cat.cache[0]

    cache_paths = cache.load(cat._urlpath, output=False)
    cache_path = cache_paths[-1]

    assert cache_path.startswith(cache._cache_dir)
    assert os.path.isfile(cache_path)

    cache.clear_all()


def test_get_metadata(catalog_cache):
    cat = catalog_cache['test_cache']
    cache = cat.cache[0]
    cache_paths = cache.load(cat._urlpath, output=False)

    metadata = cache.get_metadata(cat._urlpath)

    assert isinstance(metadata, list)
    for d in metadata:
        assert d['cache_path'] in cache_paths
        assert 'created' in d.keys()
        assert 'original_path' in d.keys()
    cache.clear_all()


def test_clear_cache(catalog_cache):
    cat = catalog_cache['test_cache']
    cache = cat.cache[0]
    cache_paths = cache.load(cat._urlpath, output=False)

    cache.clear_cache(cat._urlpath)

    assert cat._urlpath not in cache._metadata.keys()
    for cache_path in cache_paths:
        assert os.path.basename(cache_path) not in os.listdir(cache._cache_dir)


def test_clear_cache_bad_metadata(catalog_cache):
    cat = catalog_cache['test_cache']
    cache = cat.cache[0]
    cache_paths = cache.load(cat._urlpath, output=False)

    subdir = os.path.dirname(cache_paths[0])

    shutil.rmtree(subdir)

    cache.clear_cache(cat._urlpath)

    assert cat._urlpath not in cache._metadata.keys()
    for cache_path in cache_paths:
        assert os.path.basename(cache_path) not in os.listdir(cache._cache_dir)


def test_clear_all(catalog_cache):
    cat = catalog_cache['test_cache']
    cache = cat.cache[0]
    cache_paths = cache.load(cat._urlpath, output=False)

    cache.clear_all()

    for cache_path in cache_paths:
        assert not os.path.exists(cache_path)

    cache.clear_all()


def test_second_load(catalog_cache):
    cat = catalog_cache['test_cache']
    cache = cat.cache[0]

    cache_paths = cache.load(cat._urlpath, output=False)
    cache_path = cache_paths[-1]

    assert os.path.isfile(cache_path)
    t1 = os.path.getmtime(cache_path)

    cache.load(cat._urlpath, output=False)
    assert os.path.isfile(cache_path)
    t2 = os.path.getmtime(cache_path)
    assert t1 == t2

    cache.clear_all()


def test_second_load_timestamp(catalog_cache):
    cat = catalog_cache['test_cache']
    cache = cat.cache[0]

    cache_paths = cache.load(cat._urlpath, output=False)
    cache_path = cache_paths[-1]

    time1 = os.path.getmtime(cache_path)

    cache.clear_cache(cat._urlpath)
    assert not os.path.isfile(cache_path)
    time.sleep(0.5)

    cache.load(cat._urlpath, output=False)
    assert os.path.isfile(cache_path)

    time2 = os.path.getmtime(cache_path)
    assert time1 < time2

    cache.clear_all()


def test_second_load_refresh(catalog_cache):
    cat = catalog_cache['test_cache']
    cache = cat.cache[0]

    cache_paths = cache.load(cat._urlpath, output=False)
    cache_path = cache_paths[-1]

    time1 = os.path.getmtime(cache_path)

    assert os.path.isfile(cache_path)

    cache.load(cat._urlpath, output=False)
    assert os.path.isfile(cache_path)

    time2 = os.path.getmtime(cache_path)
    assert time1 == time2

    cache.clear_all()


def test_multiple_cache(catalog_cache):
    cat = catalog_cache['test_multiple_cache']

    assert len(cat.cache) == 2

    for cache in cat.cache:

        cache_paths = cache.load(cat._urlpath, output=False)
        cache_path = cache_paths[-1]

        assert cache._cache_dir in cache_path
        assert os.path.isfile(cache_path)

        cache.clear_all()


def test_disable_caching(catalog_cache):
    conf['cache_disabled'] = True

    cat = catalog_cache['test_cache']
    cache = cat.cache[0]

    cache_paths = cache.load(cat._urlpath, output=False)
    cache_path = cache_paths[-1]

    assert cache_path == cat._urlpath

    conf['cache_disabled'] = False

    cache_paths = cache.load(cat._urlpath, output=False)
    cache_path = cache_paths[-1]

    assert cache._cache_dir in cache_path
    assert os.path.isfile(cache_path)

    cache_id = os.path.basename(os.path.dirname(cache_path))
    # Checking for md5 hash
    assert all(c in string.hexdigits for c in cache_id)
    cache.clear_all()


def test_ds_set_cache_dir(catalog_cache):
    cat = catalog_cache['test_cache']()
    defaults = cat.cache_dirs

    new_cache_dir = os.path.join(os.getcwd(), 'test_cache_dir')
    cat.set_cache_dir(new_cache_dir)

    cache = cat.cache[0]
    assert make_path_posix(cache._cache_dir) == make_path_posix(new_cache_dir)

    cache_paths = cache.load(cat._urlpath, output=False)
    cache_path = cache_paths[-1]
    expected_cache_dir = make_path_posix(new_cache_dir)

    assert expected_cache_dir in cache_path
    assert defaults[0] not in cache_path
    assert os.path.isfile(cache_path)

    cache_id = os.path.basename(os.path.dirname(cache_path))
    # Checking for md5 hash
    assert all(c in string.hexdigits for c in cache_id)
    cache.clear_all()

    shutil.rmtree(expected_cache_dir)



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os
import os.path

import pytest
import requests
import shutil
import subprocess
import time

from tornado.ioloop import IOLoop
from tornado.testing import AsyncHTTPTestCase
import msgpack

from intake import open_catalog
from intake.container.serializer import MsgPackSerializer, GzipCompressor
from intake.cli.server.server import IntakeServer
from intake.compat import unpack_kwargs, pack_kwargs
from intake.utils import make_path_posix

catalog_file = make_path_posix(
    os.path.join(os.path.dirname(__file__), 'catalog1.yml'))


class TestServerV1Base(AsyncHTTPTestCase):
    def get_app(self):
        local_catalog = open_catalog(catalog_file)
        self.server = IntakeServer(local_catalog)
        return self.server.make_app()

    def encode(self, msg):
        return msgpack.packb(msg, **pack_kwargs)

    def decode(self, bytestr):
        return msgpack.unpackb(bytestr, **unpack_kwargs)


class TestServerV1Info(TestServerV1Base):
    def test_info(self):
        response = self.fetch('/v1/info')
        self.assertEqual(response.code, 200)

        info = self.decode(response.body)

        self.assertTrue('version' in info)

        expected = [
            {
                'container': 'dataframe',
                'direct_access': 'forbid',
                'description': 'example1 source plugin',
                'name': 'use_example1',
                'user_parameters': []},
            {
                'container': 'dataframe',
                'direct_access': 'forbid',
                'description': 'entry1 full',
                'name': 'entry1',
                'user_parameters': []},
            {
                'container': 'dataframe',
                'direct_access': 'allow',
                'description': 'entry1 part',
                'name': 'entry1_part',
                'user_parameters': [{
                    'allowed': ['1', '2'],
                    'default': '1',
                    'description': 'part of filename',
                    'name': 'part',
                    'type': 'str'
                }]
            }
        ]

        def sort_by_name(seq):
            return sorted(seq, key=lambda d: d['name'])

        for left, right in zip(sort_by_name(info['sources']),
                               sort_by_name(expected)):
            for k in right:
                assert left[k] == right[k]


class TestServerV1Source(TestServerV1Base):
    def make_post_request(self, msg, expected_status=200):
        request = self.encode(msg)
        response = self.fetch('/v1/source', method='POST', body=request,
                              headers={'Content-type':
                                       'application/vnd.msgpack'})
        self.assertEqual(response.code, expected_status)

        responses = []
        if expected_status < 400:
            unpacker = msgpack.Unpacker(**unpack_kwargs)
            unpacker.feed(response.body)

            for msg in unpacker:
                responses.append(msg)
        else:
            responses = [{'error': str(response.error)}]

        return responses

    def test_open(self):
        msg = dict(action='open', name='entry1', parameters={})
        resp_msg, = self.make_post_request(msg)

        self.assertEqual(resp_msg['container'], 'dataframe')
        self.assertEqual(resp_msg['shape'], [None, 3])
        expected_dtype = {'name': 'object', 'score': 'float64', 'rank': 'int64'}
        actual_dtype = resp_msg['dtype']
        self.assertEqual(expected_dtype, actual_dtype)
        self.assertEqual(resp_msg['npartitions'], 2)
        md = resp_msg['metadata']
        md.pop('catalog_dir', None)
        self.assertEqual(md, dict(foo='bar', bar=[1, 2, 3]))

        self.assertTrue(isinstance(resp_msg['source_id'], str))

    def test_open_direct(self):
        msg = dict(action='open', name='entry1_part', parameters=dict(part='2'),
                   available_plugins=['csv'])
        resp_msg, = self.make_post_request(msg)

        self.assertTrue('csv' in resp_msg['plugin'])
        args = resp_msg['args']
        assert 'urlpath' in args
        self.assertTrue(args['urlpath'].endswith('/entry1_2.csv'))
        md = resp_msg['metadata']
        md.pop('catalog_dir', None)
        self.assertEqual(md, dict(foo='baz', bar=[2, 4, 6]))
        self.assertEqual(resp_msg['description'], 'entry1 part')

    def test_read_part_compressed(self):
        # because the msgpack format actually depends on pyarrow
        pytest.importorskip('pyarrow')

        msg = dict(action='open', name='entry1', parameters={})
        resp_msg, = self.make_post_request(msg)
        source_id = resp_msg['source_id']

        msg2 = dict(action='read', source_id=source_id,
                    accepted_formats=['msgpack'], accepted_compression=['gzip'],
                    partition=0)
        resp_msgs = self.make_post_request(msg2)

        self.assertEqual(len(resp_msgs), 1)
        ser = MsgPackSerializer()
        comp = GzipCompressor()

        for chunk in resp_msgs:
            self.assertEqual(chunk['format'], 'msgpack')
            self.assertEqual(chunk['compression'], 'gzip')
            self.assertEqual(chunk['container'], 'dataframe')

            data = ser.decode(comp.decompress(chunk['data']),
                              container='dataframe')
            self.assertEqual(len(data), 4)

    def test_read_partition(self):
        # because the msgpack format actually depends on pyarrow
        pytest.importorskip('pyarrow')
        msg = dict(action='open', name='entry1', parameters={})
        resp_msg, = self.make_post_request(msg)
        source_id = resp_msg['source_id']

        msg2 = dict(action='read', partition=1, source_id=source_id,
                    accepted_formats=['msgpack'])
        resp_msgs = self.make_post_request(msg2)

        self.assertEqual(len(resp_msgs), 1)
        ser = MsgPackSerializer()

        part = resp_msgs[0]
        self.assertEqual(part['format'], 'msgpack')
        self.assertEqual(part['compression'], 'none')
        self.assertEqual(part['container'], 'dataframe')

        data = ser.decode(part['data'], container='dataframe')
        self.assertEqual(len(data), 4)

    def test_bad_action(self):
        msg = dict(action='bad', name='entry1')
        response, = self.make_post_request(msg, expected_status=400)
        self.assertIn('bad', response['error'])

    def test_no_format(self):
        msg = dict(action='open', name='entry1', parameters={})
        resp_msg, = self.make_post_request(msg)
        source_id = resp_msg['source_id']

        msg2 = dict(action='read', source_id=source_id,
                    accepted_formats=['unknown_format'])
        response, = self.make_post_request(msg2, expected_status=400)
        self.assertIn('compatible', response['error'])

    def test_idle_timer(self):
        self.server.start_periodic_functions(close_idle_after=0.1,
                                             remove_idle_after=0.2)

        msg = dict(action='open', name='entry1', parameters={})
        resp_msg, = self.make_post_request(msg)
        source_id = resp_msg['source_id']

        # Let ioloop run once with do-nothing function to make sure source
        # isn't closed
        time.sleep(0.05)
        IOLoop.current().run_sync(lambda: None)

        # Cheat and look into internal state now
        source = self.server._cache.peek(source_id)
        assert source._dataframe is not None

        # now wait slightly over idle time, run periodic functions,
        # and check again
        time.sleep(0.06)
        IOLoop.current().run_sync(lambda: None)

        # should be closed
        source = self.server._cache.peek(source_id)
        assert source._dataframe is None

        # wait a little longer
        time.sleep(0.1)
        IOLoop.current().run_sync(lambda: None)

        # source should be gone
        with self.assertRaises(KeyError):
            self.server._cache.peek(source_id)


@pytest.fixture()
def multi_server(tmpdir):
    fn1 = make_path_posix(os.path.join(tmpdir, 'cat1.yaml'))
    shutil.copyfile(catalog_file, fn1)
    fn2 = make_path_posix(os.path.join(tmpdir, 'cat2.yaml'))
    shutil.copyfile(catalog_file, fn2)
    P = subprocess.Popen(['intake-server', fn1, fn2, '--no-flatten'])
    t = time.time()
    while True:
        try:
            requests.get('http://localhost:5000')
            yield 'intake://localhost:5000'
            break
        except:
            time.sleep(0.2)
            if time.time() - t > 10:
                break
    P.terminate()
    P.wait()
    shutil.rmtree(tmpdir)


def test_flatten_flag(multi_server):
    cat = open_catalog(multi_server)
    assert list(cat) == ['cat1', 'cat2']
    assert 'use_example1' in cat.cat1()


def free_port():
    import socket

    s = socket.socket()
    s.bind(('', 0))  # Bind to a free port provided by the host.
    port = s.getsockname()[1]
    s.close()
    return port


@pytest.fixture(scope='function')
def port_server(tmpdir):
    fn1 = make_path_posix(os.path.join(tmpdir, 'cat1.yaml'))
    shutil.copyfile(catalog_file, fn1)
    port = free_port()

    P = subprocess.Popen(['intake-server', '--port', str(port), fn1])
    t = time.time()
    try:
        while True:
            try:
                requests.get('http://localhost:%s' % port)
                yield 'intake://localhost:%s' % port
                break
            except:
                time.sleep(0.2)
                if time.time() - t > 10:
                    break
    finally:
        P.terminate()
        P.wait()
        shutil.rmtree(tmpdir)


def test_port_flag(port_server):
    cat = open_catalog(port_server)
    assert 'use_example1' in list(cat)


@pytest.fixture()
def address_server(tmpdir):
    fn1 = make_path_posix(os.path.join(tmpdir, 'cat1.yaml'))
    shutil.copyfile(catalog_file, fn1)
    port = free_port()
    P = subprocess.Popen(['intake-server', '--port', str(port), '--address', '0.0.0.0', fn1])
    t = time.time()
    while True:
        try:
            requests.get('http://0.0.0.0:%s' % port)
            yield 'intake://0.0.0.0:%s' % port
            break
        except:
            time.sleep(0.2)
            if time.time() - t > 10:
                break
    P.terminate()
    P.wait()
    shutil.rmtree(tmpdir)


def test_address_flag(address_server):
    cat = open_catalog(address_server)
    assert 'use_example1' in list(cat)
